# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
from hypothesis import settings, given, strategies as st
import hypothesis.extra.numpy as hnp
from lattpy.utils import (
    SiteOccupiedError,
    NoAtomsError,
    NoConnectionsError,
    NotAnalyzedError,
)
from lattpy import Lattice, Circle, Atom
import lattpy as lp

settings.load_profile("lattpy")

atom = Atom()


PI = np.pi
TWOPI = 2 * np.pi

chain = Lattice.chain(a=1.0)
rchain = Lattice(TWOPI)

square = Lattice.square(a=1.0)
rsquare = Lattice(TWOPI * np.eye(2))

rect = Lattice.rectangular(a1=2.0, a2=1.0)
rrect = Lattice(PI * np.array([[1, 0], [0, 2]]))


hexagonal = Lattice.hexagonal(a=1)
rhexagonal = Lattice(np.array([[+2.0943951, +3.62759873], [+2.0943951, -3.62759873]]))
sc = Lattice.sc(a=1.0)
rsc = Lattice(TWOPI * np.eye(3))

fcc = Lattice.fcc(a=1.0)
rfcc = Lattice(TWOPI * np.array([[+1, +1, -1], [+1, -1, +1], [-1, +1, +1]]))
bcc = Lattice.bcc(a=1.0)
rbcc = Lattice(TWOPI * np.array([[+1, +1, 0], [0, -1, +1], [-1, 0, +1]]))

LATTICES = [chain, square, rect, hexagonal, sc, fcc, bcc]
RLATTICES = [rchain, rsquare, rrect, rhexagonal, rsc, rfcc, rbcc]

STRUCTURES = list()
_latt = lp.simple_chain()
_latt.build(4)
STRUCTURES.append(_latt)
_latt = lp.simple_square()
_latt.build((4, 4))
STRUCTURES.append(_latt)
_latt = lp.simple_cubic()
_latt.build((4, 4, 4))
STRUCTURES.append(_latt)


@st.composite
def structures(draw):
    return draw(st.sampled_from(STRUCTURES))


def assert_elements_equal1d(actual, expected):
    actual = np.unique(actual)
    expected = np.unique(expected)
    assert len(actual) == len(expected)
    return all(np.isin(actual, expected))


def assert_allclose_elements(actual, expected, atol=0.0, rtol=1e-7):
    assert_allclose(np.sort(actual), np.sort(expected), rtol, atol)


def assert_equal_elements(actual, expected):
    assert_array_equal(np.sort(actual), np.sort(expected))


def test_is_reciprocal():
    for latt, rlatt in zip(LATTICES, RLATTICES):
        rvecs = rlatt.vectors
        assert latt.is_reciprocal(rvecs)
        assert not latt.is_reciprocal(-1 * rvecs)
        assert not latt.is_reciprocal(+2 * rvecs)
        assert not latt.is_reciprocal(0.5 * rvecs)
        assert not latt.is_reciprocal(0.0 * rvecs)


def test_reciprocal_vectors():
    for latt, rlatt in zip(LATTICES, RLATTICES):
        expected = rlatt.vectors
        actual = latt.reciprocal_vectors()
        assert_allclose(expected, actual)


def test_reciprocal_vectors_double():
    for latt in LATTICES:
        expected = latt.vectors
        actual = latt.reciprocal_lattice().reciprocal_vectors()
        assert_array_equal(expected, actual)


def test_translate():
    # Square lattice
    expected = [2.0, 0.0]
    actual = square.translate([2, 0], [0.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0.0, 2.0]
    actual = square.translate([0, 2], [0.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [1.0, 2.0]
    actual = square.translate([1, 2], [0.0, 0.0])
    assert_array_equal(expected, actual)

    # Rectangular lattice
    expected = [4.0, 0.0]
    actual = rect.translate([2, 0], [0.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0.0, 2.0]
    actual = rect.translate([0, 2], [0.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [2.0, 2.0]
    actual = rect.translate([1, 2], [0.0, 0.0])
    assert_array_equal(expected, actual)


def test_itranslate():
    # Square lattice
    expected = [2, 0], [0.0, 0.0]
    actual = square.itranslate([2.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0, 2], [0.0, 0.0]
    actual = square.itranslate([0.0, 2.0])
    assert_array_equal(expected, actual)

    expected = [1, 2], [0.0, 0.0]
    actual = square.itranslate([1.0, 2.0])
    assert_array_equal(expected, actual)

    # Rectangular lattice
    expected = [1, 0], [0.0, 0.0]
    actual = rect.itranslate([2.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0, 2], [0.0, 0.0]
    actual = rect.itranslate([0.0, 2.0])
    assert_array_equal(expected, actual)

    expected = [1, 1], [0.0, 0.0]
    actual = rect.itranslate([2.0, 1.0])
    assert_array_equal(expected, actual)


def test_brillouin_zone():
    latt = Lattice.square()
    bz = latt.brillouin_zone()

    expected = [[-1.0, -1.0], [1.0, -1.0], [-1.0, 1.0], [1.0, 1.0]]
    assert_array_equal(bz.vertices / np.pi, expected)

    expected = [[0, 1], [0, 2], [1, 3], [2, 3]]
    assert_array_equal(bz.edges, expected)


@given(st.integers(1, 3))
def test_add_atom(dim):
    latt = Lattice(np.eye(dim))

    latt.add_atom()
    assert_array_equal(latt.atom_positions[0], np.zeros(dim))

    with pytest.raises(SiteOccupiedError):
        latt.add_atom()

    with pytest.raises(ValueError):
        latt.add_atom(np.zeros(4))


def test_get_alpha():
    latt = Lattice(np.eye(2))
    at1 = latt.add_atom([0.0, 0.0], atom="A")
    at2 = latt.add_atom([0.5, 0.5], atom="B")

    assert latt.get_alpha("A") == [0]
    assert latt.get_alpha(at1) == 0

    assert latt.get_alpha("B") == [1]
    assert latt.get_alpha(at2) == 1

    latt = Lattice(np.eye(2))
    latt.add_atom([0.0, 0.0], atom="A")
    latt.add_atom([0.5, 0.5], atom="A")
    assert latt.get_alpha("A") == [0, 1]


def test_get_atom():
    latt = Lattice(np.eye(2))
    at1 = latt.add_atom([0.0, 0.0], atom="A")
    at2 = latt.add_atom([0.5, 0.5], atom="B")

    assert latt.get_atom("A") == at1
    assert latt.get_atom(0) == at1

    assert latt.get_atom("B") == at2
    assert latt.get_atom(1) == at2


def test_add_connection():
    latt = Lattice(np.eye(2))
    latt.add_atom([0.0, 0.0], atom="A")
    latt.add_atom([0.5, 0.5], atom="B")

    latt.add_connection("A", "A", 1)
    latt.add_connection("A", "B", 1)
    latt.analyze()

    # Assert neighbor atom index is right
    assert all(latt.get_neighbors(alpha=0, distidx=0)[:, -1] == 1)
    assert all(latt.get_neighbors(alpha=0, distidx=1)[:, -1] == 0)
    assert all(latt.get_neighbors(alpha=1, distidx=0)[:, -1] == 0)


def test_analyze_exceptions():
    latt = Lattice(np.eye(2))
    with pytest.raises(NoAtomsError):
        latt.analyze()

    latt.add_atom()
    with pytest.raises(NoConnectionsError):
        latt.analyze()


@given(hnp.arrays(np.int64, 2, elements=st.integers(0, 10)), st.integers(0, 1))
def test_get_position(nvec, alpha):
    latt = Lattice(np.eye(2))
    latt.add_atom([0.0, 0.0], atom="A")
    latt.add_atom([0.5, 0.5], atom="B")

    pos = latt.translate(nvec, latt.atom_positions[alpha])
    assert_allclose(latt.get_position(nvec, alpha), pos)


@given(hnp.arrays(np.int64, (10, 2), elements=st.integers(0, 10)), st.integers(0, 1))
def test_get_positions(nvecs, alpha):
    latt = Lattice(np.eye(2))
    latt.add_atom([0.0, 0.0], atom="A")
    latt.add_atom([0.5, 0.5], atom="B")

    indices = np.array([[*nvec, alpha] for nvec in nvecs])
    results = latt.get_positions(indices)
    for res, nvec in zip(results, nvecs):
        pos = latt.translate(nvec, latt.atom_positions[alpha])
        assert_allclose(res, pos)


def test_estimate_index():
    # Square lattice
    expected = [2, 0]
    actual = square.estimate_index([2.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0, 2]
    actual = square.estimate_index([0.0, 2.0])
    assert_array_equal(expected, actual)

    expected = [1, 2]
    actual = square.estimate_index([1.0, 2.0])
    assert_array_equal(expected, actual)

    # Rectangular lattice
    expected = [1, 0]
    actual = rect.estimate_index([2.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0, 2]
    actual = rect.estimate_index([0.0, 2.0])
    assert_array_equal(expected, actual)

    expected = [1, 1]
    actual = rect.estimate_index([2.0, 1.0])
    assert_array_equal(expected, actual)


def test_get_neighbors():
    # chain
    latt = Lattice.chain()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1, 0], [-1, 0]])
    for idx in latt.get_neighbors(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[-2, 0], [2, 0]])
    for idx in latt.get_neighbors(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))

    # square
    latt = Lattice.square()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1, 0, 0], [0, -1, 0], [0, 1, 0], [-1, 0, 0]])
    for idx in latt.get_neighbors(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[1, -1, 0], [-1, -1, 0], [-1, 1, 0], [1, 1, 0]])
    for idx in latt.get_neighbors(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))


def test_get_neighbor_positions():
    # chain
    latt = Lattice.chain()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1], [-1]])
    for idx in latt.get_neighbor_positions(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[-2], [2]])
    for idx in latt.get_neighbor_positions(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))

    # square
    latt = Lattice.square()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1, 0], [0, -1], [0, 1], [-1, 0]])
    for idx in latt.get_neighbor_positions(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[1, -1], [-1, -1], [-1, 1], [1, 1]])
    for idx in latt.get_neighbor_positions(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))


def test_get_neighbor_vectors():
    # chain
    latt = Lattice.chain()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1], [-1]])
    for idx in latt.get_neighbor_vectors(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[-2], [2]])
    for idx in latt.get_neighbor_vectors(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))

    # square
    latt = Lattice.square()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1, 0], [0, -1], [0, 1], [-1, 0]])
    for idx in latt.get_neighbor_vectors(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[1, -1], [-1, -1], [-1, 1], [1, 1]])
    for idx in latt.get_neighbor_vectors(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))


def test_volume():
    latt = Lattice(np.eye(2))
    latt.add_atom()
    latt.add_connections()
    latt.build((4, 4), primitive=False)
    assert latt.volume() == 25


def test_alpha():
    latt = Lattice(np.eye(2))
    latt.add_atom([0.0, 0.0], "A")
    latt.add_atom([0.5, 0.5], "B")
    latt.add_connection("A", "A", 1)
    latt.add_connection("A", "B", 1)
    latt.analyze()
    latt.build((4, 4), primitive=False)

    assert latt.alpha(0) == 0
    assert latt.alpha(1) == 1
    assert latt.alpha(2) == 0
    assert latt.alpha(3) == 1


def test_atom():
    latt = Lattice(np.eye(2))
    at1 = latt.add_atom([0.0, 0.0], "A")
    at2 = latt.add_atom([0.5, 0.5], "B")
    latt.add_connection("A", "A", 1)
    latt.add_connection("A", "B", 1)
    latt.analyze()
    latt.build((4, 4), primitive=False)

    assert latt.atom(0) == at1
    assert latt.atom(1) == at2
    assert latt.atom(2) == at1
    assert latt.atom(3) == at2


def test_position():
    latt = Lattice(np.eye(2))
    latt.add_atom([0.0, 0.0], "A")
    latt.add_atom([0.5, 0.5], "B")
    latt.add_connection("A", "A", 1)
    latt.add_connection("A", "B", 1)
    latt.analyze()
    latt.build((4, 4), primitive=False)

    assert_array_equal(latt.position(0), [0.0, 0.0])
    assert_array_equal(latt.position(1), [0.5, 0.5])
    assert_array_equal(latt.position(2), [0.0, 1.0])
    assert_array_equal(latt.position(3), [0.5, 1.5])


def test_center():
    latt = lp.simple_square()
    latt.build((10, 10))
    assert_array_equal(latt.center(), [5, 5])

    latt.build((3, 3))
    assert_array_equal(latt.center(), [1.5, 1.5])

    latt.build((6, 3))
    assert_array_equal(latt.center(), [3, 1.5])


def test_center_of_gravity():
    latt = lp.simple_square()
    latt.build((10, 10))
    assert_array_equal(latt.center_of_gravity(), [5, 5])

    latt.build((3, 3))
    assert_array_equal(latt.center_of_gravity(), [1.5, 1.5])

    latt.build((6, 3))
    assert_array_equal(latt.center_of_gravity(), [3, 1.5])


def test_superindex_from_position():
    latt = Lattice(np.eye(2))
    latt.add_atom()
    latt.add_connections()
    latt.build((4, 4), primitive=False)
    for i in range(latt.num_sites):
        pos = latt.position(i)
        assert latt.superindex_from_pos(pos) == i


def test_superindex_from_index():
    latt = Lattice(np.eye(2))
    latt.add_atom()
    latt.add_connections()
    latt.build((4, 4), primitive=False)
    for i in range(latt.num_sites):
        ind = latt.indices[i]
        assert latt.superindex_from_index(ind) == i


def test_get_base_atom_dict():
    latt = Lattice(np.eye(2))
    ata = latt.add_atom([0, 0], atom="A")
    atb = latt.add_atom([0.5, 0], atom="B")
    latt.add_atom([0.5, 0.5], atom="B")
    result = latt.get_base_atom_dict()

    assert len(result) == 2
    assert_array_equal(result[ata], [[0, 0]])
    assert_array_equal(result[atb], [[0.5, 0.0], [0.5, 0.5]])


def test_build():
    pass


def test_build_min_neighbors():
    latt = Lattice(np.eye(2))
    latt.add_atom()
    latt.add_connections()
    shape = Circle([0, 0], radius=5)
    latt.build(shape, min_neighbors=2)
    assert_array_equal(latt.data.get_limits(), [[-4.0, -4.0], [4.0, 4.0]])


def test_build_exceptions():
    latt = Lattice(np.eye(2))
    latt.add_atom()

    with pytest.raises(NoConnectionsError):
        latt.build((5, 5))

    latt.add_connections(analyze=False)
    with pytest.raises(NotAnalyzedError):
        latt.build((5, 5))

    latt.analyze()
    with pytest.raises(ValueError):
        latt.build((5, 5, 5))


@given(structures())
def test_nearest_neighbors(latt):
    for i in range(latt.num_sites):
        expected = latt.neighbors(i, distidx=0)
        actual = latt.nearest_neighbors(i)
        assert_equal_elements(actual, expected)

        expected = latt.neighbors(i, distidx=0, unique=True)
        actual = latt.nearest_neighbors(i, unique=True)
        assert_equal_elements(actual, expected)


@given(structures())
def test_iter_neighbors(latt):
    for i in range(latt.num_sites):
        for distidx, actual in latt.iter_neighbors(i):
            expected = latt.neighbors(i, distidx=distidx)
            assert_equal_elements(actual, expected)
        for distidx, actual in latt.iter_neighbors(i, unique=True):
            expected = latt.neighbors(i, distidx=distidx, unique=True)
            assert_equal_elements(actual, expected)


@given(structures())
def test_check_neighbors(latt):
    for i in range(latt.num_sites):
        for distidx, neighbors in latt.iter_neighbors(i):
            # Check neighbors
            for j in neighbors:
                assert latt.check_neighbors(i, j) == distidx
            # Check not neighbors
            for j in np.random.uniform(0, latt.num_sites, size=20):
                if j not in list(neighbors):
                    assert latt.check_neighbors(i, j) is None


def test_compute_connections():
    latt = Lattice(np.eye(2))
    latt.add_atom()
    latt.add_connections()
    latt.build((4, 4), primitive=False)

    latt2 = Lattice(np.eye(2))
    latt2.add_atom()
    latt2.add_connections()
    latt2.build((4, 4), pos=(5, 0), primitive=False)

    pairs, dists = latt.compute_connections(latt2)
    expected = [[20, 0], [21, 1], [22, 2], [23, 3], [24, 4]]
    assert_array_equal(pairs, expected)
    assert np.all(dists) == 1

    latt2.build((4, 4), pos=(5, 1), primitive=False)

    pairs, dists = latt.compute_connections(latt2)
    expected = [[21, 0], [22, 1], [23, 2], [24, 3]]
    assert_array_equal(pairs, expected)
    assert np.all(dists) == 1


def test_periodic_nearest():
    # Lattice chain
    latt = lp.simple_chain()
    latt.build(9)
    latt.set_periodic(0)
    assert 9 in latt.neighbors(0)

    # Square lattice
    latt = lp.simple_square()
    latt.build((4, 4))

    latt.set_periodic(0)
    assert_elements_equal1d(latt.nearest_neighbors(0), [1, 5, 20])
    assert_elements_equal1d(latt.nearest_neighbors(1), [0, 2, 6, 21])
    assert_elements_equal1d(latt.nearest_neighbors(2), [1, 3, 7, 22])
    assert_elements_equal1d(latt.nearest_neighbors(3), [2, 4, 8, 23])
    assert_elements_equal1d(latt.nearest_neighbors(4), [3, 9, 24])
    latt.set_periodic(1)
    assert_elements_equal1d(latt.nearest_neighbors(0), [1, 5, 4])
    assert_elements_equal1d(latt.nearest_neighbors(5), [0, 6, 10, 9])
    assert_elements_equal1d(latt.nearest_neighbors(10), [5, 11, 15, 14])
    assert_elements_equal1d(latt.nearest_neighbors(15), [10, 16, 20, 19])
    assert_elements_equal1d(latt.nearest_neighbors(20), [15, 21, 24])
    # Only check corners for both axis periodic
    latt.set_periodic([0, 1])
    assert_elements_equal1d(latt.nearest_neighbors(0), [1, 5, 20, 24])
    assert_elements_equal1d(latt.nearest_neighbors(4), [3, 9, 20, 24])
    assert_elements_equal1d(latt.nearest_neighbors(20), [0, 15, 21, 24])
    assert_elements_equal1d(latt.nearest_neighbors(24), [4, 19, 20, 23])

    # graphene lattice
    latt = lp.graphene()
    latt.build((5.5, 4.5))

    latt.set_periodic(0)
    assert_elements_equal1d(latt.nearest_neighbors(0), [1, 15])
    assert_elements_equal1d(latt.nearest_neighbors(2), [3, 15, 21])
    assert_elements_equal1d(latt.nearest_neighbors(8), [9, 21, 23])
    latt.set_periodic(1)
    assert_elements_equal1d(latt.nearest_neighbors(1), [0, 4, 16])
    assert_elements_equal1d(latt.nearest_neighbors(6), [5, 7, 17])
    assert_elements_equal1d(latt.nearest_neighbors(7), [6, 14, 22])
    # Only check corners for both axis periodic
    latt.set_periodic([0, 1])
    assert_elements_equal1d(latt.nearest_neighbors(0), [1, 15, 23])
    assert_elements_equal1d(latt.nearest_neighbors(8), [9, 21, 23])


def test_periodic_next_nearest():
    # Lattice chain
    latt = lp.simple_chain(neighbors=2)
    latt.build(9)
    latt.set_periodic(0)
    assert 8 in latt.neighbors(0, distidx=1)

    # Square lattice
    latt = lp.simple_square(neighbors=2)
    latt.build((4, 4))
    latt.set_periodic(0)
    assert_elements_equal1d(latt.neighbors(0, 1), [6, 21])
    assert_elements_equal1d(latt.neighbors(1, 1), [7, 5, 20, 22])
    assert_elements_equal1d(latt.neighbors(2, 1), [8, 6, 21, 23])
    assert_elements_equal1d(latt.neighbors(3, 1), [9, 7, 22, 24])
    assert_elements_equal1d(latt.neighbors(4, 1), [8, 23])
    latt.set_periodic(1)
    assert_elements_equal1d(latt.neighbors(0, 1), [6, 9])
    assert_elements_equal1d(latt.neighbors(5, 1), [1, 4, 11, 14])
    assert_elements_equal1d(latt.neighbors(10, 1), [6, 9, 16, 19])
    assert_elements_equal1d(latt.neighbors(15, 1), [11, 14, 21, 24])
    assert_elements_equal1d(latt.neighbors(20, 1), [16, 19])
    # Only check corners for both axis periodic
    latt.set_periodic([0, 1])
    assert_elements_equal1d(latt.neighbors(0, 1), [6, 9, 21, 24])
    assert_elements_equal1d(latt.neighbors(4, 1), [5, 8, 20, 23])
    assert_elements_equal1d(latt.neighbors(20, 1), [1, 4, 16, 19])
    assert_elements_equal1d(latt.neighbors(23, 1), [0, 3, 15, 18])


def test_periodic_small():
    latt = lp.simple_square()
    latt.build((2, 2), primitive=True)
    latt.set_periodic(True)

    assert_elements_equal1d(latt.neighbors(0), [1, 2])
    assert_elements_equal1d(latt.neighbors(1), [2, 3])


def test_periodic_notbuilt():
    latt = lp.simple_square()
    with pytest.raises(lp.NotBuiltError):
        latt.set_periodic(True)


def test_remove_periodic():
    latt = lp.simple_chain()
    latt.build(9)
    latt.set_periodic(0)
    latt.set_periodic(None)
    assert 9 not in latt.neighbors(0)


def test_minimum_distances_chain():
    latt = lp.simple_chain()
    latt.build(4)
    assert_array_equal(latt.minimum_distances(0), [0.0, 1.0, 2.0, 3.0, 4.0])
    assert_array_equal(latt.minimum_distances(1), [1.0, 0.0, 1.0, 2.0, 3.0])
    assert_array_equal(latt.minimum_distances(2), [2.0, 1.0, 0.0, 1.0, 2.0])
    assert_array_equal(latt.minimum_distances(3), [3.0, 2.0, 1.0, 0.0, 1.0])
    assert_array_equal(latt.minimum_distances(4), [4.0, 3.0, 2.0, 1.0, 0.0])
    latt.set_periodic(0)
    assert_array_equal(latt.minimum_distances(0), [0.0, 1.0, 2.0, 2.0, 1.0])
    assert_array_equal(latt.minimum_distances(1), [1.0, 0.0, 1.0, 2.0, 2.0])
    assert_array_equal(latt.minimum_distances(2), [2.0, 1.0, 0.0, 1.0, 2.0])
    assert_array_equal(latt.minimum_distances(3), [2.0, 2.0, 1.0, 0.0, 1.0])
    assert_array_equal(latt.minimum_distances(4), [1.0, 2.0, 2.0, 1.0, 0.0])


def test_minimum_distances_square():
    latt = lp.simple_square()
    latt.build((2, 2))
    sqrt11 = np.sqrt(2)
    sqrt21 = np.sqrt(5)
    sqrt22 = np.sqrt(8)

    # lower left
    dists = [0.0, 1.0, 2.0, 1.0, sqrt11, sqrt21, 2.0, sqrt21, sqrt22]
    assert_allclose(latt.minimum_distances(0), dists)
    # upper left
    dists = [2.0, 1.0, 0.0, sqrt21, sqrt11, 1.0, sqrt22, sqrt21, 2.0]
    assert_allclose(latt.minimum_distances(2), dists)
    # center
    dists = [sqrt11, 1.0, sqrt11, 1.0, 0.0, 1, sqrt11, 1.0, sqrt11]
    assert_allclose(latt.minimum_distances(4), dists)
    # lower right
    dists = [2.0, sqrt21, sqrt22, 1.0, sqrt11, sqrt21, 0.0, 1.0, 2.0]
    assert_allclose(latt.minimum_distances(6), dists)
    # upper right
    dists = [sqrt22, sqrt21, 2.0, sqrt21, sqrt11, 1.0, 2.0, 1.0, 0.0]
    assert_allclose(latt.minimum_distances(8), dists)

    latt.set_periodic(0)
    # lower left
    dists = [0.0, 1.0, 2.0, 1.0, sqrt11, sqrt21, 1.0, sqrt11, sqrt21]
    assert_allclose(latt.minimum_distances(0), dists)
    # upper left
    dists = [2.0, 1.0, 0.0, sqrt21, sqrt11, 1.0, sqrt21, sqrt11, 1.0]
    assert_allclose(latt.minimum_distances(2), dists)
    # center
    dists = [sqrt11, 1.0, sqrt11, 1.0, 0.0, 1, sqrt11, 1.0, sqrt11]
    assert_allclose(latt.minimum_distances(4), dists)
    # lower right
    dists = [1.0, sqrt11, sqrt21, 1.0, sqrt11, sqrt21, 0.0, 1.0, 2.0]
    assert_allclose(latt.minimum_distances(6), dists)
    # upper right
    dists = [sqrt21, sqrt11, 1.0, sqrt21, sqrt11, 1.0, 2.0, 1.0, 0.0]
    assert_allclose(latt.minimum_distances(8), dists)


def test_append():
    latt = Lattice(np.eye(2))
    latt.add_atom()
    latt.add_connections()
    latt.build((4, 4), primitive=False)
    latt2 = latt.copy()
    latt.append(latt2)

    assert_elements_equal1d(latt.nearest_neighbors(20), [15, 21, 25])
    assert_elements_equal1d(latt.nearest_neighbors(21), [16, 20, 22, 26])
    assert_elements_equal1d(latt.nearest_neighbors(22), [17, 21, 23, 27])
    assert_elements_equal1d(latt.nearest_neighbors(23), [18, 22, 24, 28])
    assert_elements_equal1d(latt.nearest_neighbors(24), [19, 23, 29])


def test_extend():
    # Only check size, connections are handled by append
    latt = lp.simple_chain()
    latt.build(4, primitive=False)
    latt.extend(2)
    assert latt.num_sites == 8

    latt = lp.simple_square()
    latt.build((4, 4), primitive=False)
    latt.extend(2, ax=0)
    assert_array_equal(latt.shape, (7, 4))

    latt = lp.simple_square()
    latt.build((4, 4), primitive=False)
    latt.extend(2, ax=1)
    assert_array_equal(latt.shape, (4, 7))


def test_repeat():
    # Only check size, connections are handled by append
    latt = lp.simple_chain()
    latt.build(4, primitive=False)
    latt.repeat()
    assert latt.num_sites == 10

    latt = lp.simple_square()
    latt.build((4, 4), primitive=False)
    latt.repeat(1, ax=0)
    assert_array_equal(latt.shape, (9, 4))

    latt = lp.simple_square()
    latt.build((4, 4), primitive=False)
    latt.repeat(1, ax=1)
    assert_array_equal(latt.shape, (4, 9))


def test_to_dict():
    latt = Lattice(np.eye(2))
    latt.add_atom()
    latt.add_connections()
    latt.build((5, 5))
    d = latt.todict()

    expected = [[1.0, 0.0], [0.0, 1.0]]
    assert_array_equal(d["vectors"], expected)

    expected = [[0.0, 0.0]]
    assert_array_equal(d["positions"], expected)

    expected = [[1]]
    assert_array_equal(d["connections"], expected)

    expected = [5.0, 5.0]
    assert_array_equal(d["shape"], expected)


@given(structures())
def test_fromdict(latt):
    d = latt.todict()
    latt2 = Lattice.fromdict(d)

    assert_array_equal(latt.vectors, latt2.vectors)
    assert_array_equal(latt.atom_positions, latt2.atom_positions)
    assert_array_equal(latt._connections, latt2._connections)
    assert_array_equal(latt.num_neighbors, latt2.num_neighbors)


def test_hash():
    latt = Lattice(np.eye(2))
    latt.add_atom(atom=atom)
    latt.add_connections()
    latt.build((5, 5))
    hash1 = latt.__hash__()
    latt.build((6, 5))
    assert latt.__hash__() != hash1


@given(structures())
def test_neighbor_pairs(latt: lp.Lattice):
    pairs, distindices = latt.neighbor_pairs()
    for (i, j), distidx in zip(pairs, distindices):  # noqa
        neighbors = latt.neighbors(i, distidx=distidx)
        assert j in neighbors


@given(structures())
def test_neighbor_pairs_unique(latt: lp.Lattice):
    pairs, distindices = latt.neighbor_pairs(unique=True)
    for (i, j), distidx in zip(pairs, distindices):  # noqa
        assert i < j
        neighbors = latt.neighbors(i, distidx=distidx)
        assert j in neighbors


@given(structures())
def test_adjacency_matrix(latt: lp.Lattice):
    adjmat = latt.adjacency_matrix()
    for i in range(latt.num_sites):
        for distidx, neighbors in latt.iter_neighbors(i):
            for j in neighbors:
                assert adjmat[i, j] == distidx + 1


# =========================================================================
# General tests
# =========================================================================


def test_simple_chain():
    latt = lp.simple_chain(a=1, neighbors=1)
    latt.build(3)

    # Check correct building
    assert latt.num_sites == 4

    expected = np.atleast_2d([0.0, 1.0, 2.0, 3.0]).T
    actual = latt.data.positions
    assert_array_equal(expected, actual)

    latt = lp.simple_chain(a=1, neighbors=1)
    latt.build(3)
    # Check neighbors
    expected = [1]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, actual)

    expected = [0, 2]
    actual = latt.nearest_neighbors(1)
    assert_array_equal(expected, sorted(actual))

    # Check periodic boundary conditions
    latt.set_periodic(0)
    expected = [1, 3]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    latt = lp.simple_chain(a=1, neighbors=2)
    latt.build(6)
    expected = [0, 2, 3]
    actual = latt.neighbors(1)
    assert_array_equal(expected, sorted(actual))

    expected = [0, 1, 3, 4]
    actual = latt.neighbors(2)
    assert_array_equal(expected, sorted(actual))


def test_simple_square():
    latt = lp.simple_square(a=1, neighbors=1)
    latt.build((2, 2))

    # Check correct building
    assert latt.num_sites == 9

    # Check nearest neighbors
    expected = [1, 3, 5, 7]
    actual = latt.nearest_neighbors(4)
    assert_array_equal(expected, sorted(actual))

    # Check periodic boundary conditions
    latt.set_periodic(0)
    expected = [1, 3, 6]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    expected = [0, 2, 4, 7]
    actual = latt.nearest_neighbors(1)
    assert_array_equal(expected, sorted(actual))

    latt.set_periodic(1)
    expected = [1, 2, 3]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    # Check next nearest neighbors
    latt = lp.simple_square(a=1, neighbors=2)
    latt.build((2, 2))

    expected = [0, 2, 6, 8]
    actual = latt.neighbors(4, distidx=1)
    assert_array_equal(expected, sorted(actual))


def test_simple_cubic():
    latt = lp.simple_cubic(a=1, neighbors=1)
    latt.build((2, 2, 2))

    # Check correct building
    assert latt.num_sites == 27

    # Check nearest neighbors
    expected = [4, 10, 12, 14, 16, 22]
    actual = latt.nearest_neighbors(13)
    assert_array_equal(expected, sorted(actual))

    expected = [1, 3, 9]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    # Check periodic boundary conditions
    latt.set_periodic(0)
    expected = [1, 3, 9, 18]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    latt.set_periodic(1)
    expected = [1, 3, 6, 9]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    latt.set_periodic(2)
    expected = [1, 2, 3, 9]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    latt.set_periodic([0, 1])
    expected = [1, 3, 6, 9, 18]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    latt.set_periodic([1, 2])
    expected = [1, 2, 3, 6, 9]
    actual = latt.nearest_neighbors(0)
    assert_array_equal(expected, sorted(actual))

    # Check next nearest neighbors
    latt = lp.simple_cubic(a=1, neighbors=2)
    latt.build((2, 2, 2))

    expected = [1, 3, 5, 7, 9, 11, 15, 17, 19, 21, 23, 25]
    actual = latt.neighbors(13, distidx=1)
    assert_array_equal(expected, sorted(actual))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
import math

sys.path.insert(0, os.path.abspath("../.."))

import lattpy


# -- Project information -----------------------------------------------------

project = "lattpy"
copyright = "2022, Dylan L. Jones"
author = "Dylan L. Jones"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "numpydoc",
    "myst_parser",
    "sphinx.ext.autosummary",
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.intersphinx",  # links to numpy, scipy ... docs
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.extlinks",  # define roles for links
    "sphinx_toggleprompt",  # toggle `>>>`
    "sphinx_rtd_theme",
    "sphinx.ext.graphviz",
    "sphinx.ext.inheritance_diagram",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.

exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "tests",
    "generated/lattpy.rst",
    "generated/modules.rst",
]

# -- Graphviz options --------------------------------------------------------

graphviz_output_format = "svg"
inheritance_graph_attrs = dict(rankdir="LR", size='""')


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Include Markdown parser
# source_parsers = {
#    '.md': 'recommonmark.parser.CommonMarkParser',
# }
source_suffix = [".rst", ".md"]

# Don't show type hints
autodoc_typehints = "none"

# Preserve order
autodoc_member_order = "bysource"

# -- Apidoc ------------------------------------------------------------------

add_module_names = True


# -- Autosummary -------------------------------------------------------------

autosummary_generate = True
# autosummary_imported_members = True

# -- Numpy extension ---------------------------------------------------------

numpydoc_use_plots = True
# numpydoc_xref_param_type = True
# numpydoc_xref_ignore = "all"  # not working...
numpydoc_show_class_members = False

# -- Plots -------------------------------------------------------------------

plot_pre_code = """
import warnings
import numpy as np
import matplotlib.pyplot as plt
import lattpy as lp
np.random.seed(0)
"""

doctest_global_setup = plot_pre_code  # make doctests consistent
doctest_global_cleanup = """
try:
    plt.close()  # close any open figures
except:
    pass
"""

plot_include_source = True
plot_html_show_source_link = False
plot_formats = [("png", 100), "pdf"]

phi = (math.sqrt(5) + 1) / 2

plot_rcparams = {
    "font.size": 8,
    "axes.titlesize": 8,
    "axes.labelsize": 8,
    "xtick.labelsize": 8,
    "ytick.labelsize": 8,
    "legend.fontsize": 8,
    "figure.figsize": (3 * phi, 3),
    "figure.subplot.bottom": 0.2,
    "figure.subplot.left": 0.2,
    "figure.subplot.right": 0.9,
    "figure.subplot.top": 0.85,
    "figure.subplot.wspace": 0.4,
    "text.usetex": False,
}

# -- Intersphinx -------------------------------------------------------------

# taken from https://gist.github.com/bskinn/0e164963428d4b51017cebdb6cda5209
intersphinx_mapping = {
    "python": (r"https://docs.python.org", None),
    "scipy": (r"https://docs.scipy.org/doc/scipy/", None),
    "numpy": (r"https://docs.scipy.org/doc/numpy/", None),
    "np": (r"https://docs.scipy.org/doc/numpy/", None),
    "matplotlib": (r"https://matplotlib.org/", None),
}


# -- Auto-run sphinx-apidoc --------------------------------------------------


def run_apidoc(_):
    from sphinx.ext.apidoc import main
    import os
    import sys

    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    proj_dir = os.path.dirname(os.path.dirname(cur_dir))
    doc_dir = os.path.join(proj_dir, "docs")
    output_path = os.path.join(doc_dir, "source", "generated")
    module = os.path.join(proj_dir, "lattpy")
    exclude = os.path.join(module, "tests")
    template_dir = os.path.join(doc_dir, "source", "_templates", "apidoc")
    main(["-fMeT", "-o", output_path, module, exclude, "--templatedir", template_dir])


def setup(app):
    app.connect("builder-inited", run_apidoc)



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Basis object for defining the coordinate system and unit cell of a lattice."""

import logging
import itertools
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D
from typing import Union, Sequence, Callable
from .spatial import cell_size, cell_volume, WignerSeitzCell
from .plotting import subplot, draw_unit_cell

logger = logging.getLogger(__name__)

vecs_t = Union[float, Sequence[float], Sequence[Sequence[float]]]
basis_t = Union[float, Sequence[float], Sequence[Sequence[float]], "LatticeBasis"]

__all__ = ["basis_t", "LatticeBasis"]


class LatticeBasis:
    """Lattice basis for representing the coordinate system and unit cell of a lattice.

    The ``LatticeBasis`` object is the core of any lattice model. It defines the
    basis vectors and subsequently the coordinate system of the lattice and provides
    the necessary basis transformations between the world and lattice coordinate system.

    Attributes
    ----------
    dim
    vectors
    vectors3d
    norms
    cell_size
    cell_volume

    Methods
    -------
    chain
    square
    rectangular
    oblique
    hexagonal
    hexagonal3d
    sc
    fcc
    bcc
    itransform
    transform
    itranslate
    translate
    is_reciprocal
    reciprocal_vectors
    reciprocal_lattice
    get_neighbor_cells
    wigner_seitz_cell
    brillouin_zone
    plot_basis

    Parameters
    ----------
    basis: array_like or float or LatticeBasis
        The primitive basis vectors that define the unit cell of the lattice. If a
        ``LatticeBasis`` instance is passed it is copied and used as the new basis.
    **kwargs
        Key-word arguments. Used only when subclassing ``LatticeBasis``.

    Examples
    --------
    >>> import lattpy as lp
    >>> import matplotlib.pyplot as plt
    >>> basis = lp.LatticeBasis.square()
    >>> _ = basis.plot_basis()
    >>> plt.show()

    """

    # Tolerance for reciprocal vectors/lattice
    RVEC_TOLERANCE: float = 1e-6

    # noinspection PyUnusedLocal
    def __init__(self, basis: basis_t, **kwargs):
        if isinstance(basis, LatticeBasis):
            basis = basis.vectors
        # Vector basis
        self._vectors = np.atleast_2d(basis).T
        self._vectors_inv = np.linalg.inv(self._vectors)
        self._dim = len(self._vectors)
        self._cell_size = cell_size(self.vectors)
        self._cell_volume = cell_volume(self.vectors)

    @classmethod
    def chain(cls, a: float = 1.0, **kwargs):
        """Initializes a one-dimensional lattice."""
        return cls(a, **kwargs)

    @classmethod
    def square(cls, a: float = 1.0, **kwargs):
        """Initializes a 2D lattice with square basis vectors."""
        return cls(a * np.eye(2), **kwargs)

    @classmethod
    def rectangular(cls, a1: float = 1.0, a2: float = 1.0, **kwargs):
        """Initializes a 2D lattice with rectangular basis vectors."""
        return cls(np.array([[a1, 0], [0, a2]]), **kwargs)

    @classmethod
    def oblique(cls, alpha: float, a1: float = 1.0, a2: float = 1.0, **kwargs):
        """Initializes a 2D lattice with oblique basis vectors."""
        vectors = np.array([[a1, 0], [a2 * np.cos(alpha), a2 * np.sin(alpha)]])
        return cls(vectors, **kwargs)

    @classmethod
    def hexagonal(cls, a: float = 1.0, **kwargs):
        """Initializes a 2D lattice with hexagonal basis vectors."""
        vectors = a / 2 * np.array([[3, np.sqrt(3)], [3, -np.sqrt(3)]])
        return cls(vectors, **kwargs)

    @classmethod
    def hexagonal3d(cls, a: float = 1.0, az: float = 1.0, **kwargs):
        """Initializes a 3D lattice with hexagonal basis vectors."""
        vectors = (
            a / 2 * np.array([[3, np.sqrt(3), 0], [3, -np.sqrt(3), 0], [0, 0, az]])
        )
        return cls(vectors, **kwargs)

    @classmethod
    def sc(cls, a: float = 1.0, **kwargs):
        """Initializes a 3D simple cubic lattice."""
        return cls(a * np.eye(3), **kwargs)

    @classmethod
    def fcc(cls, a: float = 1.0, **kwargs):
        """Initializes a 3D face centered cubic lattice."""
        vectors = a / 2 * np.array([[1, 1, 0], [1, 0, 1], [0, 1, 1]])
        return cls(vectors, **kwargs)

    @classmethod
    def bcc(cls, a: float = 1.0, **kwargs):
        """Initializes a 3D body centered cubic lattice."""
        vectors = a / 2 * np.array([[1, 1, 1], [1, -1, 1], [-1, 1, 1]])
        return cls(vectors, **kwargs)

    @classmethod
    def hypercubic(cls, dim, a: float = 1.0, **kwargs):
        """Creates a d-dimensional cubic lattice."""
        vectors = a * np.eye(dim)
        return cls(vectors, **kwargs)

    # ==================================================================================

    @property
    def dim(self) -> int:
        """int: The dimension of the lattice."""
        return self._dim

    @property
    def vectors(self) -> np.ndarray:
        """np.ndarray: Array containing the basis vectors as rows."""
        return self._vectors.T

    @property
    def vectors3d(self) -> np.ndarray:
        """np.ndarray: The basis vectors expanded to three dimensions."""
        vectors = np.eye(3)
        vectors[: self.dim, : self.dim] = self._vectors
        return vectors.T

    @property
    def norms(self) -> np.ndarray:
        """np.ndarray: Lengths of the basis vectors."""
        return np.linalg.norm(self._vectors, axis=0)

    @property
    def cell_size(self) -> np.ndarray:
        """np.ndarray: The shape of the box spawned by the basis vectors."""
        return self._cell_size

    @property
    def cell_volume(self) -> float:
        """float: The volume of the unit cell defined by the basis vectors."""
        return self._cell_volume

    def itransform(
        self, world_coords: Union[Sequence[int], Sequence[Sequence[int]]]
    ) -> np.ndarray:
        """Transform the world coords ``(x, y, ...)`` into the basis coords
        ``(n, m, ...)``.

        Parameters
        ----------
        world_coords : (..., N) array_like
            The coordinates in the world coordinate system that are transformed
            into the lattice coordinate system.

        Returns
        -------
        basis_coords : (..., N) np.ndarray
            The coordinates in the lattice coordinate system.

        Examples
        --------
        Construct a lattice with basis vectors :math:`a_1 = (2, 0)` and
        :math:`a_2 = (0, 1)`:

        >>> latt = LatticeBasis([[2, 0], [0, 1]])

        Transform points into the coordinate system of the lattice:

        >>> latt.itransform([2, 0])
        [1. 0.]

        >>> latt.itransform([4, 0])
        [2. 0.]

        >>> latt.itransform([0, 1])
        [0. 1.]
        """
        world_coords = np.atleast_1d(world_coords)
        return np.inner(world_coords, self._vectors_inv)

    def transform(
        self, basis_coords: Union[Sequence[int], Sequence[Sequence[int]]]
    ) -> np.ndarray:
        """Transform the basis-coords ``(n, m, ...)`` into the world coords
        ``(x, y, ...)``.

        Parameters
        ----------
        basis_coords : (..., N) array_like
            The coordinates in the lattice coordinate system that are transformed
            into the world coordinate system.

        Returns
        -------
        world_coords : (..., N) np.ndarray
            The coordinates in the cartesian coordinate system.

        Examples
        --------
        Construct a lattice with basis vectors :math:`a_1 = (2, 0)` and
        :math:`a_2 = (0, 1)` :

        >>> latt = LatticeBasis([[2, 0], [0, 1]])

        Transform points into the world coordinat system:

        >>> latt.itransform([1, 0])
        [2. 0.]

        >>> latt.itransform([2, 0])
        [4. 0.]

        >>> latt.itransform([0, 1])
        [0. 1.]
        """
        basis_coords = np.atleast_1d(basis_coords)
        return np.inner(basis_coords, self._vectors)

    def translate(
        self,
        nvec: Union[int, Sequence[int], Sequence[Sequence[int]]],
        r: Union[float, Sequence[float]] = 0.0,
    ) -> np.ndarray:
        r"""Translates the given postion vector ``r`` by the translation vector ``n``.

        The position is calculated using the translation vector :math:`n` and the
        atom position in the unitcell :math:`r`:

        .. math::
            R = \sum_i n_i v_i + r

        Parameters
        ----------
        nvec : (..., N) array_like
            Translation vector in the lattice coordinate system.
        r : (N) array_like, optional
            The position in cartesian coordinates. If no vector is passed only
            the translation is returned.

        Returns
        -------
        r_trans : (..., N) np.ndarray
            The translated position.

        Examples
        --------
        Construct a lattice with basis vectors :math:`a_1 = (2, 0)` and
        :math:`a_2 = (0, 1)`:

        >>> latt = LatticeBasis([[2, 0], [0, 1]])

        Translate the origin:

        >>> n = [1, 0]
        >>> latt.translate(n)
        [2. 0.]

        Translate a point:

        >>> p = [0.5, 0.5]
        >>> latt.translate(n, p)
        [2.5 0.5]

        Translate a point by multiple translation vectors:

        >>> p = [0.5, 0.5]
        >>> nvecs = [[0, 0], [1, 0], [2, 0]]
        >>> latt.translate(nvecs, p)
        [[0.5 0.5]
         [2.5 0.5]
         [4.5 0.5]]
        """
        r = np.atleast_1d(r)
        nvec = np.atleast_1d(nvec)
        return r + np.inner(nvec, self._vectors)

    def itranslate(self, x: Union[float, Sequence[float]]) -> [np.ndarray, np.ndarray]:
        """Returns the translation vector and atom position of the given position.

        Parameters
        ----------
        x : (..., N) array_like or float
            Position vector in cartesian coordinates.

        Returns
        -------
        nvec : (..., N) np.ndarray
            Translation vector in the lattice basis.
        r : (..., N) np.ndarray, optional
            The position in real-space.

        Examples
        --------
        Construct a lattice with basis vectors :math:`a_1 = (2, 0)` and
        :math:`a_2 = (0, 1)`:

        >>> latt = LatticeBasis([[2, 0], [0, 1]])
        >>> latt.itranslate([2, 0])
        (array([1., 0.]), array([0., 0.]))

        >>> latt.itranslate([2.5, 0.5])
        (array([1., 0.]), array([0.5, 0.5]))
        """
        x = np.atleast_1d(x)
        itrans = self.itransform(x)
        nvec = np.floor(itrans).astype(np.int64)
        r = x - self.translate(nvec)
        return nvec, r

    def is_reciprocal(self, vecs: vecs_t, tol: float = RVEC_TOLERANCE) -> bool:
        r"""Checks if the given vectors are reciprocal to the lattice vectors.

        The lattice- and reciprocal vectors :math:`a_i` and :math:`b_i` must satisfy
        the relation

        .. math::
            a_i \cdot b_i = 2 \pi \delta_{ij}

        To check the given vectors, the difference of each dot-product is compared to
        :math:`2\pi` with the given tolerance.

        Parameters
        ----------
        vecs : array_like or float
            The vectors to check. Must have the same dimension as the lattice.
        tol : float, optional
            The tolerance used for checking the result of the dot-products.

        Returns
        -------
        is_reciprocal : bool
            Flag if the vectors are reciprocal to the lattice basis vectors.
        """
        vecs = np.atleast_2d(vecs)
        two_pi = 2 * np.pi
        for a, b in zip(self.vectors, vecs):
            if abs(np.dot(a, b) - two_pi) > tol:
                return False
        return True

    def reciprocal_vectors(
        self, tol: float = RVEC_TOLERANCE, check: bool = False
    ) -> np.ndarray:
        r"""Computes the reciprocal basis vectors of the bravais lattice.

        The lattice- and reciprocal vectors :math:`a_i` and :math:`b_i` must satisfy
        the relation

        .. math::
            a_i \cdot b_i = 2 \pi \delta_{ij}

        Parameters
        ----------
        tol : float, optional
            The tolerance used for checking the result of the dot-products.
        check : bool, optional
            Check the result and raise an exception if it does not satisfy
            the definition.

        Returns
        -------
        v_rec : np.ndarray
            The reciprocal basis vectors of the lattice.

        Examples
        --------
        Reciprocal vectors of the square lattice:

        >>> latt = LatticeBasis(np.eye(2))
        >>> latt.reciprocal_vectors()
        [[6.28318531 0.        ]
         [0.         6.28318531]]
        """
        two_pi = 2 * np.pi
        if self.dim == 1:
            return np.array([[two_pi / self.vectors[0, 0]]])

        # Convert basis vectors of the bravais lattice to 3D, compute
        # reciprocal vectors and convert back to actual dimension.
        a1, a2, a3 = self.vectors3d
        factor = 2 * np.pi / self.cell_volume
        b1 = np.cross(a2, a3)
        b2 = np.cross(a3, a1)
        b3 = np.cross(a1, a2)
        rvecs = factor * np.asarray([b1, b2, b3])
        rvecs = rvecs[: self.dim, : self.dim]

        # Fix the sign so that the dot-products are all positive
        # and raise an exception if anything went wrong
        vecs = self.vectors
        for i in range(self.dim):
            dot = np.dot(vecs[i], rvecs[i])
            # Check if dot product is - 2 pi
            if abs(dot + two_pi) <= tol:
                rvecs[i] *= -1
            # Raise an exception if checks are enabled and
            # dot product results in anything other than +2 pi
            elif check and abs(dot - two_pi) > tol:
                raise ValueError(f"{rvecs[i]} not a reciprocal vector to {vecs[i]}")

        return rvecs

    # noinspection PyShadowingNames
    def reciprocal_lattice(self, min_negative: bool = False):
        """Creates the lattice in reciprocal space.

        Parameters
        ----------
        min_negative : bool, optional
            If True the reciprocal vectors are scaled such that
            there are fewer negative elements than positive ones.

        Returns
        -------
        rlatt : LatticeBasis
            The lattice in reciprocal space

        See Also
        --------
        reciprocal_vectors : Constructs the reciprocal vectors used for the
            reciprocal lattice

        Examples
        --------
        Reciprocal lattice of the square lattice:

        >>> latt = LatticeBasis(np.eye(2))
        >>> rlatt = latt.reciprocal_lattice()
        >>> rlatt.vectors
        [[6.28318531 0.        ]
         [0.         6.28318531]]
        """
        rvecs = self.reciprocal_vectors(min_negative)
        rlatt = self.__class__(rvecs)
        return rlatt

    def get_neighbor_cells(
        self,
        distidx: int = 0,
        include_origin: bool = True,
        comparison: Callable = np.isclose,
    ) -> np.ndarray:
        """Find all neighboring unit cells of the unit cell at the origin.

        Parameters
        ----------
        distidx : int, default
            Index of distance to neighboring cells, default is 0 (nearest neighbors).
        include_origin : bool, optional
            If True the origin is included in the set.
        comparison : callable, optional
            The method used for comparing distances.

        Returns
        -------
        indices : np.ndarray
            The lattice indeices of the neighboring unit cells.

        Examples
        --------
        >>> latt = LatticeBasis(np.eye(2))
        >>> latt.get_neighbor_cells(distidx=0, include_origin=False)
        [[-1  0]
         [ 0 -1]
         [ 0  1]
         [ 1  0]]
        """
        # Build cell points
        max_factor = distidx + 1
        axis_factors = np.arange(-max_factor, max_factor + 1)
        factors = np.array(list(itertools.product(axis_factors, repeat=self.dim)))
        points = np.dot(factors, self.vectors[np.newaxis, :, :])[:, 0, :]

        # Compute distances to origin for all points
        distances = np.linalg.norm(points, axis=1)

        # Set maximum distances value to number of neighbors
        # + number of unique vector lengths
        max_distidx = distidx + len(np.unique(np.linalg.norm(self.vectors, axis=1)))

        # Return points with distance lower than maximum distance
        maxdist = np.sort(np.unique(distances))[max_distidx]
        indices = np.where(comparison(distances, maxdist))[0]
        factors = factors[indices]

        origin = np.zeros(self.dim)
        idx = np.where((factors == origin).all(axis=1))[0]
        if include_origin and not len(idx):
            factors = np.append(origin[np.newaxis, :], factors, axis=0)
        elif not include_origin and len(idx):
            factors = np.delete(factors, idx, axis=0)
        return factors

    def wigner_seitz_cell(self) -> WignerSeitzCell:
        """Computes the Wigner-Seitz cell of the lattice structure.

        Returns
        -------
        ws_cell : WignerSeitzCell
            The Wigner-Seitz cell of the lattice.
        """
        nvecs = self.get_neighbor_cells(include_origin=True)
        positions = np.dot(nvecs, self.vectors[np.newaxis, :, :])[:, 0, :]
        return WignerSeitzCell(positions)

    def brillouin_zone(self, min_negative: bool = False) -> WignerSeitzCell:
        """Computes the first Brillouin-zone of the lattice structure.

        Constructs the Wigner-Seitz cell of the reciprocal lattice

        Parameters
        ----------
        min_negative : bool, optional
            If True the reciprocal vectors are scaled such that
            there are fewer negative elements than positive ones.

        Returns
        -------
        ws_cell : WignerSeitzCell
            The Wigner-Seitz cell of the reciprocal lattice.
        """
        rvecs = self.reciprocal_vectors(min_negative)
        rlatt = self.__class__(rvecs)
        return rlatt.wigner_seitz_cell()

    def get_cell_superindex(self, index, shape):
        """Converts a cell index to a super-index for the given shape.

        The index of a unit cell is the translation vector.

        Parameters
        ----------
        index : array_like or (N, D) np.ndarray
            One or multiple cell indices. If a numpy array is passed the result will
            be an array of super-indices, otherwise an integer is returned.
        shape : (D, ) array_like
            The shape for converting the index.

        Returns
        -------
        super_index : int or (N, ) int np.ndarray
            The super index. If `index` is a numpy array, `super_index` is an array
            of super indices, otherwise it is an integer.
        """
        single = not isinstance(index, np.ndarray)
        index = np.atleast_2d(index)

        super_indices = np.zeros((len(index)), dtype=np.int64)
        for i, ind in enumerate(index):
            si = sum(ind[d] * np.prod(shape[d + 1 :]) for d in range(self.dim))
            super_indices[i] = si

        return super_indices[0] if single else super_indices

    def get_cell_index(self, super_index, shape):
        """Converts a super index to the corresponding cell index for the given shape.

        The index of a unit cell is the translation vector.

        Parameters
        ----------
        super_index : int or (N,) int np.ndarray
            One or multiple super indices.
        shape : (D, ) array_like
            The shape for converting the index.

        Returns
        -------
        index : np.ndarray
            The cell index. Of shape (D,) if `super_index` is an integer, otherwise
            of shape (N, D).
        """
        single = isinstance(super_index, int)
        super_index = np.atleast_1d(super_index)

        indices = np.zeros((len(super_index), self.dim), dtype=np.int64)
        for i, si in enumerate(super_index):
            rest = si
            for d in range(self.dim - 1):
                val, rest = divmod(rest, np.prod(shape[d + 1 :]))
                indices[i, d] = val
            indices[i, self.dim - 1] = rest

        return indices[0] if single else indices

    def plot_basis(
        self,
        lw: float = None,
        ls: str = "--",
        margins: Union[Sequence[float], float] = 0.1,
        grid: bool = False,
        show_cell: bool = True,
        show_vecs: bool = True,
        adjustable: str = "box",
        ax: Union[plt.Axes, Axes3D] = None,
        show: bool = False,
    ) -> Union[plt.Axes, Axes3D]:  # pragma: no cover
        """Plot the lattice basis.

        Parameters
        ----------
        lw : float, optional
            The line width used for plotting the unit cell outlines.
        ls : str, optional
            The line style used for plotting the unit cell outlines.
        margins : Sequence[float] or float, optional
            The margins of the plot.
        grid : bool, optional
            If True, draw a grid in the plot.
        show_vecs : bool, optional
            If True the first unit-cell is drawn.
        show_cell : bool, optional
            If True the outlines of the unit cell are plotted.
        adjustable : None or {'box', 'datalim'}, optional
            If not None, this defines which parameter will be adjusted to meet
            the equal aspect ratio. If 'box', change the physical dimensions of
            the Axes. If 'datalim', change the x or y data limits.
            Only applied to 2D plots.
        ax : plt.Axes or plt.Axes3D or None, optional
            Parent plot. If None, a new plot is initialized.
        show : bool, optional
            If True, show the resulting plot.
        """
        logger.debug("Plotting unit cell")
        if self.dim > 3:
            raise ValueError(f"Plotting in {self.dim} dimensions is not supported!")

        hopz, atomz = range(2)

        fig, ax = subplot(self.dim, adjustable, ax)
        # Draw unit vectors and the cell they spawn.
        if show_vecs:
            vectors = self.vectors
            draw_unit_cell(ax, vectors, show_cell, lw=lw, color="k", ls=ls, zorder=hopz)
        # Configure grid
        if grid and self.dim < 3:
            ax.set_axisbelow(True)
            ax.grid(b=True, which="major")
        # Adjust margin
        if isinstance(margins, float):
            margins = [margins] * self.dim
        ax.margins(*margins)

        fig.tight_layout()
        if show:
            plt.show()
        return ax

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dim: {self.dim})"



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

import math
import numpy as np
from pytest import mark
from numpy.testing import assert_array_equal, assert_allclose
from hypothesis import given, strategies as st
import hypothesis.extra.numpy as hnp
from lattpy import spatial, simple_chain, simple_square, simple_cubic


finite_floats = st.floats(-1e6, +1e6, allow_nan=False, allow_infinity=False)


@given(
    hnp.arrays(np.float64, 10, elements=finite_floats),
    hnp.arrays(np.float64, 10, elements=finite_floats),
)
def test_distance(a, b):
    expected = math.sqrt(np.sum(np.square(a - b)))

    res = spatial.distance(a, b)
    assert res == expected

    res = spatial.distance(a, b, decimals=3)
    assert res == round(expected, 3)


@given(
    hnp.arrays(np.float64, (5, 10), elements=finite_floats),
    hnp.arrays(np.float64, (5, 10), elements=finite_floats),
)
def test_distances(a, b):
    results = spatial.distances(a, b)
    for i in range(len(results)):
        expected = math.sqrt(np.sum(np.square(a[i] - b[i])))
        assert results[i] == expected

    results = spatial.distances(a, b, 3)
    for i in range(len(results)):
        expected = math.sqrt(np.sum(np.square(a[i] - b[i])))
        assert results[i] == np.round(expected, 3)


@mark.parametrize(
    "arrays, result",
    [
        (([1, 3], [2, 4]), [1, 2, 3, 4]),
        (([1, 4], [2, 5], [3, 6]), [1, 2, 3, 4, 5, 6]),
        (([[1, 1], [3, 3]], [[2, 2], [4, 4]]), [[1, 1], [2, 2], [3, 3], [4, 4]]),
    ],
)
def test_interweave(arrays, result):
    assert_array_equal(spatial.interweave(np.array(arrays)), result)


@mark.parametrize(
    "limits, result",
    [
        (([0, 1],), [[0]]),
        (([0, 1], [0, 1]), [[0, 0]]),
        (([0, 2],), [[0], [1]]),
        (([0, 2], [0, 1]), [[0, 0], [1, 0]]),
        (([0, 2], [0, 2]), [[0, 0], [0, 1], [1, 0], [1, 1]]),
        (([0, 3], [0, 2]), [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]]),
    ],
)
def test_vindices(limits, result):
    assert_array_equal(spatial.vindices(limits), result)


@mark.parametrize(
    "stop, result",
    [
        (1, [[0]]),
        (3, [[0], [1], [2]]),
        ((3, 2), [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]]),
    ],
)
def test_vrange_stop(stop, result):
    assert_array_equal(spatial.vrange(stop), result)


@mark.parametrize(
    "start, stop, result",
    [
        (0, 1, [[0]]),
        (0, 3, [[0], [1], [2]]),
        (1, 3, [[1], [2]]),
        ((0, 0), (3, 2), [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]]),
        ((1, 0), (3, 2), [[1, 0], [1, 1], [2, 0], [2, 1]]),
    ],
)
def test_vrange_startstop(start, stop, result):
    assert_array_equal(spatial.vrange(start, stop), result)


@mark.parametrize(
    "vecs, result",
    [
        ([[1, 0], [0, 1]], [1.0, 1.0]),
        ([[2, 0], [0, 1]], [2.0, 1.0]),
        ([[2, 0], [1, 1]], [2.0, 1.0]),
        ([[2, 0], [-1, 1]], [3.0, 1.0]),
        ([[2, 0, 0], [0, 2, 0], [0, 0, 2]], [2.0, 2.0, 2.0]),
    ],
)
def test_cell_size(vecs, result):
    assert_array_equal(spatial.cell_size(vecs), result)


@mark.parametrize(
    "vecs, result",
    [
        ([[1, 0], [0, 1]], 1.0),
        ([[2, 0], [0, 1]], 2.0),
        ([[2, 0], [0, 2]], 4.0),
        ([[2, 0, 0], [0, 2, 0], [0, 0, 2]], 8.0),
    ],
)
def test_cell_colume(vecs, result):
    assert spatial.cell_volume(vecs) == result


def test_wignerseitz_symmetry_points():
    # Test 1D
    latt = simple_chain()
    ws = latt.wigner_seitz_cell()
    origin, corners, edge_centers, face_centers = ws.symmetry_points()
    assert_array_equal(origin, [0.0])
    assert_array_equal(corners, [[-0.5], [0.5]])
    assert edge_centers is None
    assert face_centers is None

    # Test 2D
    latt = simple_square()
    ws = latt.wigner_seitz_cell()
    origin, corners, edge_centers, face_centers = ws.symmetry_points()

    assert_array_equal(origin, [0.0, 0.0])
    assert_array_equal(
        2 * corners, [[-1.0, -1.0], [1.0, -1.0], [-1.0, 1.0], [1.0, 1.0]]
    )
    assert_array_equal(
        2 * edge_centers, [[0.0, -1.0], [-1.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
    )
    assert face_centers is None

    # Test 3D
    latt = simple_cubic()
    ws = latt.wigner_seitz_cell()
    origin, corners, edge_centers, face_centers = ws.symmetry_points()

    c = [
        [-0.5, -0.5, -0.5],
        [0.5, -0.5, -0.5],
        [-0.5, -0.5, 0.5],
        [0.5, -0.5, 0.5],
        [-0.5, 0.5, -0.5],
        [0.5, 0.5, -0.5],
        [0.5, 0.5, 0.5],
        [-0.5, 0.5, 0.5],
    ]
    e = [
        [-0.5, -0.5, 0.0],
        [0.0, -0.5, -0.5],
        [0.5, -0.5, 0.0],
        [0.0, -0.5, 0.5],
        [-0.5, 0.5, 0.0],
        [0.0, 0.5, 0.5],
        [0.5, 0.5, 0.0],
        [0.0, 0.5, -0.5],
        [0.5, 0.0, 0.5],
        [0.0, 0.5, 0.5],
        [-0.5, 0.0, 0.5],
        [0.0, -0.5, 0.5],
        [-0.5, -0.5, 0.0],
        [-0.5, 0.0, -0.5],
        [-0.5, 0.5, 0.0],
        [-0.5, 0.0, 0.5],
        [0.5, 0.0, -0.5],
        [0.0, 0.5, -0.5],
        [-0.5, 0.0, -0.5],
        [0.0, -0.5, -0.5],
        [0.5, -0.5, 0.0],
        [0.5, 0.0, -0.5],
        [0.5, 0.5, 0.0],
        [0.5, 0.0, 0.5],
    ]

    f = [
        [0.0, -0.5, 0.0],
        [0.0, 0.5, 0.0],
        [0.0, 0.0, 0.5],
        [-0.5, 0.0, 0.0],
        [0.0, 0.0, -0.5],
        [0.5, 0.0, 0.0],
    ]

    assert_array_equal(origin, [0.0, 0.0, 0.0])
    assert_array_equal(corners, c)
    assert_array_equal(edge_centers, e)
    assert_array_equal(face_centers, f)


def test_wigner_seitz_arange():
    latt = simple_square()
    ws = latt.wigner_seitz_cell()

    x, y = ws.arange(0.1)
    assert_allclose(x, np.arange(-0.5, +0.5, 0.1))
    assert_allclose(y, np.arange(-0.5, +0.5, 0.1))

    x, y = ws.arange((0.1, 0.2))
    assert_allclose(x, np.arange(-0.5, +0.5, 0.1))
    assert_allclose(y, np.arange(-0.5, +0.5, 0.2))


def test_wigner_seitz_linspace():
    latt = simple_square()
    ws = latt.wigner_seitz_cell()

    x, y = ws.linspace(100, endpoint=True)
    assert_allclose(x, np.linspace(-0.5, +0.5, 100))
    assert_allclose(y, np.linspace(-0.5, +0.5, 100))

    x, y = ws.linspace((100, 200), endpoint=True)
    assert_allclose(x, np.linspace(-0.5, +0.5, 100))
    assert_allclose(y, np.linspace(-0.5, +0.5, 200))


def test_wigner_seitz_meshgrid():
    latt = simple_square()
    ws = latt.wigner_seitz_cell()

    num = 101
    actual = ws.meshgrid(num, endpoint=True)
    x = np.linspace(-0.5, +0.5, num)
    y = np.linspace(-0.5, +0.5, num)
    expected = np.array(np.meshgrid(x, y))
    assert_allclose(actual, expected)

    step = 0.1
    actual = ws.meshgrid(steps=step, endpoint=True)
    x = np.arange(-0.5, +0.5, step)
    y = np.arange(-0.5, +0.5, step)
    expected = np.array(np.meshgrid(x, y))
    assert_allclose(actual, expected)


def test_compute_vectors():
    # Test square vectors
    vecs = spatial.compute_vectors(1.0, 1.0, alpha=90)
    assert_allclose(vecs, np.eye(2), atol=1e-16)

    # Text hexagonal vectors
    vecs = spatial.compute_vectors(1.0, 1.0, alpha=60)
    expected = np.array([[1, 0], [0.5, math.sqrt(3) / 2]])
    assert_allclose(vecs, expected, atol=1e-16)

    # Test cubic vectors
    vecs = spatial.compute_vectors(1.0, 1.0, 1.0, alpha=90, beta=90, gamma=90)
    assert_allclose(vecs, np.eye(3), atol=1e-16)


def test_rx():
    expected = np.eye(3)
    assert_allclose(spatial.rx(0), expected)

    expected = [
        [1.0, 0.0, 0.0],
        [0.0, 0.70710678, -0.70710678],
        [0.0, 0.70710678, 0.70710678],
    ]
    assert_allclose(spatial.rx(np.pi / 4), expected)

    expected = [[1, 0, 0], [0, 0, -1], [0, 1, 0]]
    assert_allclose(spatial.rx(np.pi / 2), expected, atol=1e-10)

    expected = [[1, 0, 0], [0, -1, 0], [0, 0, -1]]
    assert_allclose(spatial.rx(np.pi), expected, atol=1e-10)


def test_ry():
    expected = np.eye(3)
    assert_allclose(spatial.ry(0), expected)

    expected = [
        [0.70710678, 0.0, 0.70710678],
        [0.0, 1.0, 0.0],
        [-0.70710678, 0.0, 0.70710678],
    ]
    assert_allclose(spatial.ry(np.pi / 4), expected)

    expected = np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
    assert_allclose(spatial.ry(np.pi / 2), expected, atol=1e-10)

    expected = [[-1, 0, 0], [0, 1, 0], [0, 0, -1]]
    assert_allclose(spatial.ry(np.pi), expected, atol=1e-10)


def test_rz():
    expected = np.eye(3)
    assert_allclose(spatial.rz(0), expected)

    expected = [
        [0.70710678, -0.70710678, 0.0],
        [0.70710678, 0.70710678, 0.0],
        [0.0, 0.0, 1.0],
    ]
    assert_allclose(spatial.rz(np.pi / 4), expected)

    expected = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
    assert_allclose(spatial.rz(np.pi / 2), expected, atol=1e-10)

    expected = [[-1, 0, 0], [0, -1, 0], [0, 0, 1]]
    assert_allclose(spatial.rz(np.pi), expected, atol=1e-10)



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

from pytest import mark
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
from lattpy.basis import LatticeBasis

PI = np.pi
TWOPI = 2 * np.pi

chain = LatticeBasis.chain(a=1.0)
rchain = LatticeBasis(TWOPI)

square = LatticeBasis.square(a=1.0)
rsquare = LatticeBasis(TWOPI * np.eye(2))

rect = LatticeBasis.rectangular(a1=2.0, a2=1.0)
rrect = LatticeBasis(PI * np.array([[1, 0], [0, 2]]))


hexagonal = LatticeBasis.hexagonal(a=1)
rhexagonal = LatticeBasis(
    np.array([[+2.0943951, +3.62759873], [+2.0943951, -3.62759873]])
)
sc = LatticeBasis.sc(a=1.0)
rsc = LatticeBasis(TWOPI * np.eye(3))

fcc = LatticeBasis.fcc(a=1.0)
rfcc = LatticeBasis(TWOPI * np.array([[+1, +1, -1], [+1, -1, +1], [-1, +1, +1]]))
bcc = LatticeBasis.bcc(a=1.0)
rbcc = LatticeBasis(TWOPI * np.array([[+1, +1, 0], [0, -1, +1], [-1, 0, +1]]))

LATTICES = [chain, square, rect, hexagonal, sc, fcc, bcc]
RLATTICES = [rchain, rsquare, rrect, rhexagonal, rsc, rfcc, rbcc]


def assert_elements_equal1d(actual, expected):
    actual = np.unique(actual)
    expected = np.unique(expected)
    assert len(actual) == len(expected)
    return all(np.isin(actual, expected))


def assert_allclose_elements(actual, expected, atol=0.0, rtol=1e-7):
    assert_allclose(np.sort(actual), np.sort(expected), rtol, atol)


def assert_equal_elements(actual, expected):
    assert_array_equal(np.sort(actual), np.sort(expected))


def test_is_reciprocal():
    for latt, rlatt in zip(LATTICES, RLATTICES):
        rvecs = rlatt.vectors
        assert latt.is_reciprocal(rvecs)
        assert not latt.is_reciprocal(-1 * rvecs)
        assert not latt.is_reciprocal(+2 * rvecs)
        assert not latt.is_reciprocal(0.5 * rvecs)
        assert not latt.is_reciprocal(0.0 * rvecs)


def test_reciprocal_vectors():
    for latt, rlatt in zip(LATTICES, RLATTICES):
        expected = rlatt.vectors
        actual = latt.reciprocal_vectors()
        assert_allclose(expected, actual)


def test_reciprocal_vectors_double():
    for latt in LATTICES:
        expected = latt.vectors
        actual = latt.reciprocal_lattice().reciprocal_vectors()
        assert_array_equal(expected, actual)


def test_translate():
    # Square lattice
    expected = [2.0, 0.0]
    actual = square.translate([2, 0], [0.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0.0, 2.0]
    actual = square.translate([0, 2], [0.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [1.0, 2.0]
    actual = square.translate([1, 2], [0.0, 0.0])
    assert_array_equal(expected, actual)

    # Rectangular lattice
    expected = [4.0, 0.0]
    actual = rect.translate([2, 0], [0.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0.0, 2.0]
    actual = rect.translate([0, 2], [0.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [2.0, 2.0]
    actual = rect.translate([1, 2], [0.0, 0.0])
    assert_array_equal(expected, actual)


def test_itranslate():
    # Square lattice
    expected = [2, 0], [0.0, 0.0]
    actual = square.itranslate([2.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0, 2], [0.0, 0.0]
    actual = square.itranslate([0.0, 2.0])
    assert_array_equal(expected, actual)

    expected = [1, 2], [0.0, 0.0]
    actual = square.itranslate([1.0, 2.0])
    assert_array_equal(expected, actual)

    # Rectangular lattice
    expected = [1, 0], [0.0, 0.0]
    actual = rect.itranslate([2.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0, 2], [0.0, 0.0]
    actual = rect.itranslate([0.0, 2.0])
    assert_array_equal(expected, actual)

    expected = [1, 1], [0.0, 0.0]
    actual = rect.itranslate([2.0, 1.0])
    assert_array_equal(expected, actual)


def test_brillouin_zone():
    latt = LatticeBasis.square()
    bz = latt.brillouin_zone()

    expected = [[-1.0, -1.0], [1.0, -1.0], [-1.0, 1.0], [1.0, 1.0]]
    assert_array_equal(bz.vertices / np.pi, expected)

    expected = [[0, 1], [0, 2], [1, 3], [2, 3]]
    assert_array_equal(bz.edges, expected)


@mark.parametrize("shape", [(10,), (10, 20), (10, 20, 30)])
def test_index_superindex_conversion(shape):
    shape = np.array(shape)
    latt = LatticeBasis.hypercubic(len(shape))

    num_cells = np.prod(shape)
    ind1 = np.arange(num_cells, dtype=np.int64)
    for i in ind1:
        index = latt.get_cell_index(i, shape)
        i2 = latt.get_cell_superindex(index, shape)
        assert i == i2

    indices = latt.get_cell_index(ind1, shape)
    ind2 = latt.get_cell_superindex(indices, shape)
    assert_allclose(ind2, ind1)



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones


# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Lattice structure object for defining the atom basis and neighbor connections."""

import pickle
import logging
import itertools
import collections
import numpy as np
from copy import deepcopy
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D
from typing import Union, Tuple, List, Sequence, Callable, Any, Dict
from .utils import (
    ArrayLike,
    SiteOccupiedError,
    NoAtomsError,
    NoConnectionsError,
    NotAnalyzedError,
)
from .plotting import (
    subplot,
    draw_sites,
    draw_vectors,
    draw_unit_cell,
    connection_color_array,
)
from .spatial import vindices, interweave, KDTree
from .atom import Atom
from .shape import AbstractShape
from .basis import basis_t, LatticeBasis

logger = logging.getLogger(__name__)

__all__ = ["LatticeStructure"]


class LatticeStructure(LatticeBasis):
    """Structure object representing a infinite Bravais lattice.

    Combines the ``LatticeBasis`` with a set of atoms and connections between
    them to define a general lattice structure.

    .. rubric:: Inheritance

    .. inheritance-diagram:: LatticeStructure
       :parts: 1


    Parameters
    ----------
    basis : array_like or float or LatticeBasis
        The primitive basis vectors that define the unit cell of the lattice. If a
        ``LatticeBasis`` instance is passed it is copied and used as the new basis
        of the lattice.
    **kwargs
        Key-word arguments. Used for quickly configuring a ``LatticeStructure``
        instance. Allowed keywords are:

        Properties:
        atoms: Dictionary containing the atoms to add to the lattice.
        cons: Dictionary containing the connections to add to the lattice.

    Examples
    --------
    Two dimensional lattice with one atom in the unit cell and nearest neighbors

    >>> import lattpy as lp
    >>> latt = lp.LatticeStructure(np.eye(2))
    >>> latt.add_atom()
    >>> latt.add_connections(1)
    >>> latt
    LatticeStructure(dim: 2, num_base: 1, num_neighbors: [4])

    Quick-setup of the same lattice:

    >>> import lattpy as lp
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.LatticeStructure.square(atoms={(0.0, 0.0): "A"}, cons={("A", "A"): 1})
    >>> _ = latt.plot_cell()
    >>> plt.show()

    """

    # Decimals used for rounding distances
    DIST_DECIMALS: int = 6

    def __init__(self, basis: basis_t, **kwargs):
        super().__init__(basis)
        # Atom data
        self._num_base = 0
        self._atoms = list()
        self._positions = list()

        # Raw neighbor data without including connections
        self._raw_base_neighbors = None
        self._raw_distance_matrix = None
        self._raw_num_neighbors = None

        # Neighbour data
        self._connections = None
        self._base_neighbors = None
        self._num_neighbors = None
        self._distance_matrix = None
        self._distances = None

        logger.debug(
            "LatticeStructure initialized (D=%i)\nvectors:\n%s", self.dim, self._vectors
        )

        if "atoms" in kwargs:
            atom_dict = kwargs["atoms"]
            for pos, atom in atom_dict.items():
                self.add_atom(pos, atom)
        if "cons" in kwargs:
            cons = kwargs["cons"]
            if isinstance(cons, int):
                self.add_connections(cons)
            else:
                cons_dict = kwargs["cons"]
                for pair, num in cons_dict.items():
                    self.add_connection(*pair, num)
                self.analyze()

    @property
    def num_base(self) -> int:
        """int: The number of atoms in the unit cell."""
        return self._num_base

    @property
    def atoms(self) -> List[Atom]:
        """list of Atom: List of the atoms in the unit cell."""
        return self._atoms

    @property
    def atom_positions(self) -> List[np.ndarray]:
        """list of np.ndarray: List of positions of the atoms in the unit cell."""
        return self._positions

    @property
    def num_distances(self) -> int:
        """int: The maximal number of distances between the lattice sites."""
        return int(np.max(self._connections))

    @property
    def num_neighbors(self) -> np.ndarray:
        """np.ndarray: The number of neighbors of each atom in the unitcell."""
        return self._num_neighbors

    @property
    def base_neighbors(self) -> np.ndarray:
        """np.ndarray: The neighbors of the unitcell at the origin."""
        return self._base_neighbors

    @property
    def distances(self) -> np.ndarray:
        """np.ndarray: List of distances between the lattice sites."""
        return self._distances

    def add_atom(
        self,
        pos: Union[float, Sequence[float]] = None,
        atom: Union[str, Dict[str, Any], Atom] = None,
        primitive: bool = False,
        neighbors: int = 0,
        **kwargs,
    ) -> Atom:
        """Adds a site to the basis of the lattice unit cell.

        Parameters
        ----------
        pos : (N) array_like or float, optional
            Position of site in the unit-cell. The default is the origin of the cell.
            The size of the array has to match the dimension of the lattice.
        atom : str or dict or Atom, optional
            Identifier of the site. If a string is passed, a new Atom instance is
            created.
        primitive : bool, optional
            Flag if the specified position is in cartesian or lattice coordinates.
            If True the passed position will be multiplied with the lattice vectors.
            The default is ``False`` (cartesian coordinates).
        neighbors : int, optional
            The number of neighbor distance to calculate. If the number is 0 the
            distances have to be calculated manually after configuring the
            lattice basis.
        **kwargs
            Keyword arguments for ´Atom´ constructor. Only used if a new ``Atom``
            instance is created.

        Returns
        -------
        atom : Atom
            The ``Atom``-instance of the newly added site.

        Raises
        ------
        ValueError
            Raised if the dimension of the position does not match the dimension
            of the lattice.
        ConfigurationError
            Raised if the position of the new atom is already occupied.

        Examples
        --------
        Construct a square lattice

        >>> latt = LatticeStructure(np.eye(2))

        Create an atom and add it to the origin of the unit cell of the lattice

        >>> atom1 = Atom(name="A")
        >>> latt.add_atom([0.0, 0.0], atom=atom1)
        >>> latt.get_atom(0)
        Atom(A, size=10, 0)

        An ``Atom`` instance can also be created by passing the name of the (new) atom
        and optional keyword arguments for the constructor:

        >>> latt.add_atom([0.5, 0.5], atom="B", size=15)
        >>> latt.get_atom(1)
        Atom(B, size=15, 1)
        """
        pos = np.zeros(self.dim) if pos is None else np.atleast_1d(pos)
        if primitive:
            pos = self.translate(pos)

        if len(pos) != self._dim:
            raise ValueError(
                f"Shape of the position {pos} doesn't match "
                f"the dimension {self.dim} of the lattice!"
            )
        if any(np.all(pos == x) for x in self._positions):
            raise SiteOccupiedError(atom, pos)

        if not isinstance(atom, Atom):
            atom = Atom(atom, **kwargs)

        self._atoms.append(atom)
        self._positions.append(np.asarray(pos))

        # Update number of base atoms if data is valid
        num_base = len(self._atoms)
        assert num_base == len(self._positions)
        self._num_base = num_base
        logger.debug("Added atom %s at %s", atom, pos)

        # Initial array for number of neighbour distances
        # for current number of atoms in the unit cell
        self._connections = np.zeros((num_base, num_base), dtype=np.int64)

        if neighbors:
            self.add_connections(neighbors)
        return atom

    def get_alpha(self, atom: Union[int, str, Atom]) -> Union[int, List[int]]:
        """Returns the index of the atom in the unit-cell.

        Parameters
        ----------
        atom : int or str or Atom
            The argument for getting the atom. If a ``int`` is passed
            it is interpreted as the index, if a ``str`` is passed as
            the name of an atom.

        Returns
        -------
        alpha : int or list of int
            The atom indices. If a string was passed multiple atoms with the same name
            can be returned as list.

        Examples
        --------
        Construct a lattice with two identical atoms and a third atom in the unit cell:

        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom([0, 0], atom="A")
        >>> latt.add_atom([0.5, 0], atom="B")
        >>> latt.add_atom([0.5, 0.5], atom="B")

        Get the atom index of atom A:

        >>> latt.get_alpha("A")
        [0]

        Get the indices of the atoms B:

        >>> latt.get_alpha("B")
        [1, 2]

        Since there are two atoms B in the unit cell of the lattice both indices
        are returned.
        """
        if isinstance(atom, Atom):
            return self._atoms.index(atom)
        elif isinstance(atom, str):
            return [i for i, at in enumerate(self._atoms) if atom == at.name]
        return atom

    def get_atom(self, atom: Union[int, str, Atom]) -> Atom:
        """Returns the ``Atom`` instance of the given atom in the unit cell.

        Parameters
        ----------
        atom : int or str or Atom
            The argument for getting the atom. If a ``int`` is passed
            it is interpreted as the index, if a ``str`` is passed as
            the name of an atom.

        Returns
        -------
        atom : Atom
            The ``Atom`` instance of the given site.

        See Also
        --------
        get_alpha : Get the index of the given atom.

        Examples
        --------
        Construct a lattice with one atom in the unit cell

        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom([0, 0], atom="A")

        Get the atom instance by the name

        >>> latt.get_atom("A")
        Atom(A, size=10, 0)

        or by the index:

        >>> latt.get_atom(0)
        Atom(A, size=10, 0)

        """
        if isinstance(atom, Atom):
            return atom
        elif isinstance(atom, int):
            return self._atoms[atom]
        else:
            for at in self._atoms:
                if atom == at.name:
                    return at
            raise ValueError(f"No Atom with the name '{atom}' found!")

    def add_connection(
        self,
        atom1: Union[int, str, Atom],
        atom2: Union[int, str, Atom],
        num_distances=1,
        analyze: bool = False,
    ) -> None:
        """Sets the number of distances for a specific connection between two atoms.

        Parameters
        ----------
        atom1 : int or str or Atom
            The first atom of the connected pair.
        atom2 : int or str or Atom
            The second atom of the connected pair.
        num_distances : int, optional
            The number of neighbor-distance levels, e.g. setting to ``1`` means
            only nearest neighbors. The default are nearest neighbor connections.
        analyze : bool, optional
            If True the lattice basis is analyzed after adding connections.
            If ``False`` the ``analyze``-method needs to be called manually.
            The default is ``False``.

        See Also
        --------
        add_connections : Set up connections for all atoms in the unit cell in one call.
        analyze : Called after setting up all the lattice connections.

        Examples
        --------
        Construct a square lattice with two atoms, A and B, in the unit cell:

        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom([0.0, 0.0], atom="A")
        >>> latt.add_atom([0.5, 0.5], atom="B")

        Set next nearest and nearest neighbors between the A atoms:

        >>> latt.add_connection("A", "A", num_distances=2)

        Set nearest neighbors between A and B:

        >>> latt.add_connection("A", "B", num_distances=1)

        Set nearest neighbors between the B atoms:

        >>> latt.add_connection("B", "B", num_distances=1)
        """
        alpha1 = np.atleast_1d(self.get_alpha(atom1))
        alpha2 = np.atleast_1d(self.get_alpha(atom2))
        for i, j in itertools.product(alpha1, alpha2):
            self._connections[i, j] = num_distances
            self._connections[j, i] = num_distances
        if analyze:
            self.analyze()

    def add_connections(self, num_distances=1, analyze: bool = True) -> None:
        """Sets the number of distances for all possible atom-pairs of the unitcell.

        Parameters
        ----------
        num_distances : int, optional
            The number of neighbor-distance levels, e.g. setting to ``1`` means
            only nearest neighbors. The default are nearest neighbor connections.
        analyze : bool, optional
            If True the lattice basis is analyzed after adding connections.
            If ``False`` the ``analyze``-method needs to be called manually.
            The default is True.

        See Also
        --------
        add_connection : Set up connection between two specific atoms in the unit cell.
        analyze : Called after setting up all the lattice connections.

        Examples
        --------
        Construct a square lattice with one atom in the unit cell:

        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom()

        Set nearest neighbor hopping:

        >>> latt.add_connections(num_distances=1)
        """
        self._connections.fill(num_distances)
        if analyze:
            self.analyze()

    def _compute_base_neighbors(self, max_distidx, num_jobs=1):
        logger.debug("Building indices of neighbor-cells")

        # Build indices of neighbor-cells
        self._positions = np.asarray(self._positions)
        cell_range = 2 * max_distidx
        logger.debug("Max. distidx: %i, Cell-range: %i", max_distidx, cell_range)

        nvecs = self.get_neighbor_cells(
            cell_range, include_origin=True, comparison=np.less_equal
        )
        arrays = [
            np.c_[nvecs, i * np.ones(nvecs.shape[0])] for i in range(self.num_base)
        ]
        cols = self.dim + 1
        indices = np.ravel(arrays, order="F").astype(np.int64)
        indices = indices.reshape(cols, int(indices.shape[0] / cols)).T

        # Compute positions and initialize tree
        nvecs, alphas = indices[:, :-1], indices[:, -1]
        positions = self.translate(nvecs, np.array(self.atom_positions)[alphas])
        tree = KDTree(positions, k=len(positions))

        # Compute neighbors for each distance level in the cell range
        n = self.num_base
        neighbor_array = [
            [collections.OrderedDict() for _ in range(n)] for _ in range(n)
        ]
        for a1 in range(self.num_base):
            pos = self.atom_positions[a1]
            neighbors, distances = tree.query(
                pos, num_jobs, self.DIST_DECIMALS, include_zero=True, compact=False
            )
            neighbor_indices = indices[neighbors]
            # Store neighbors of certain distance for each atom pair in the unit cell
            for dist, idx in zip(distances, neighbor_indices):
                a2 = idx[-1]
                if dist:
                    neighbor_array[a1][a2].setdefault(dist, list()).append(idx)

        # Remove extra neighbors
        for a1, a2 in itertools.product(range(n), repeat=2):
            neighbors = neighbor_array[a1][a2]
            dists = list(sorted(neighbors.keys()))
            for dist in dists[:max_distidx]:
                neighbor_array[a1][a2][dist] = np.array(neighbors[dist])
            for dist in dists[max_distidx:]:
                del neighbor_array[a1][a2][dist]

        return neighbor_array

    def _analyze_raw(self, max_distidx):
        """Analyzes the structure of the raw lattice (without connections)."""
        n = self.num_base
        # Compute raw neighbors of unit cell
        neighbor_array = self._compute_base_neighbors(max_distidx)

        # Compute the raw distance matrix and the raw number of neighbors
        raw_distance_matrix = [[list() for _ in range(n)] for _ in range(n)]
        raw_num_neighbors = np.zeros((n, n), dtype=np.int64)
        for a1, a2 in itertools.product(range(n), repeat=2):
            neighbors = neighbor_array[a1][a2]
            raw_distance_matrix[a1][a2] += list(neighbors.keys())
            raw_num_neighbors[a1, a2] = sum(len(x) for x in neighbors.values())

        # Save raw neighbor data of the unit cell
        self._raw_base_neighbors = neighbor_array
        self._raw_distance_matrix = raw_distance_matrix
        self._raw_num_neighbors = raw_num_neighbors
        logger.debug("Number of raw neighbors:\n%s", raw_num_neighbors)
        logger.debug("Raw distance-matrix:\n%s", raw_distance_matrix)

    def _assert_atoms(self):
        if len(self._atoms) == 0:
            raise NoAtomsError()

    def _assert_connections(self):
        if np.all(self._connections == 0):
            raise NoConnectionsError()

    def _assert_analyzed(self):
        if not self._base_neighbors:
            raise NotAnalyzedError()

    def analyze(self) -> None:
        """Analyzes the structure of the lattice and stores neighbor data.

        Check's distances between all sites of the bravais lattice and saves the ``n``
        lowest values. The neighbor lattice-indices of the unit-cell are also stored
        for later use. This speeds up many calculations like finding nearest neighbors.

        Raises
        ------
        NoAtomsError
            Raised if no atoms where added to the lattice. The atoms in the unit cell
            are needed for computing the neighbors and distances of the lattice.
        NoConnectionsError
            Raised if no connections have been set up.

        Notes
        -----
        Before calling the ``analyze`` function all connections in the lattice have to
        be set up.

        Examples
        --------
        Construct a square lattice with one atom in the unit cell and nearest neighbors:

        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom()
        >>> latt.add_connections(num_distances=1)

        Call `analyze` after setting up the connections

        >>> latt.analyze()
        """
        logger.debug("Analyzing lattice")

        self._assert_atoms()
        self._assert_connections()

        max_distidx = int(np.max(self._connections))
        n = self.num_base

        # Analyze the raw lattice
        self._analyze_raw(max_distidx)

        # Filter base neighbor data for configured connections and
        # store neighbors and distances as list for each atom
        base_neighbors = [collections.OrderedDict() for _ in range(n)]
        base_distance_matrix = [[list() for _ in range(n)] for _ in range(n)]
        unique_distances = set()
        for a1, a2 in itertools.product(range(n), repeat=2):
            neighbors = self._raw_base_neighbors[a1][a2]
            dists = list(neighbors.keys())
            max_dist = self._connections[a1, a2]
            for distidx, dist in enumerate(dists[:max_dist]):
                unique_distances.add(dist)
                base_neighbors[a1].setdefault(dist, list()).extend(neighbors[dist])
                base_distance_matrix[a1][a2].append(dist)
            base_distance_matrix[a1][a2] = list(sorted(base_distance_matrix[a1][a2]))

        # Convert base neighbors back to np.ndarray
        for a1 in range(self.num_base):
            for key, vals in base_neighbors[a1].items():
                base_neighbors[a1][key] = np.asarray(vals)

        max_num_distances = len(unique_distances)

        # Compute number of neighbors for each atom in the unit cell
        num_neighbors = np.zeros(self.num_base, dtype=np.int8)
        for i, neighbors in enumerate(base_neighbors):
            num_neighbors[i] = sum(len(indices) for indices in neighbors.values())

        # store distance values / keys:
        distances = np.zeros((self.num_base, max_num_distances))
        for alpha in range(self.num_base):
            try:
                dists = list(base_neighbors[alpha].keys())
            except ValueError:
                dists = list()
            distances[alpha, : len(dists)] = sorted(dists)

        self._base_neighbors = base_neighbors
        self._distance_matrix = base_distance_matrix
        self._num_neighbors = num_neighbors
        self._distances = distances
        logger.debug("Number of neighbors:\n%s", num_neighbors)
        logger.debug("Distance-matrix:\n%s", base_distance_matrix)
        logger.debug("Distances:\n%s", distances)

    def get_position(
        self, nvec: Union[int, Sequence[int]] = None, alpha: int = 0
    ) -> np.ndarray:
        """Returns the position for a given translation vector ``nvec`` and atom
        ``alpha``.

        Parameters
        ----------
        nvec : (N) array_like or int
            The translation vector.
        alpha : int, optional
            The atom index, default is 0.

        Returns
        -------
        pos : (N) np.ndarray
            The position of the transformed lattice site.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom()
        >>> latt.add_connections(1, analyze=True)
        >>> latt.get_position([1, 0], alpha=0)
        [1. 0.]
        """
        r = self._positions[alpha]
        if nvec is None:
            return r
        n = np.atleast_1d(nvec)
        return r + (self._vectors @ n)  # self.translate(n, r)

    def get_positions(self, indices):
        """Returns the positions for multiple lattice indices.

        Parameters
        ----------
        indices : (N, D+1) array_like or int
            List of lattice indices in the format :math:`(n_1, ..., n_d, α)`.

        Returns
        -------
        pos : (N, D) np.ndarray
            The positions of the lattice sites.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom()
        >>> latt.add_connections(1)
        >>> ind = [[0, 0, 0], [1, 0, 0], [1, 1, 0]]
        >>> latt.get_positions(ind)
        [[0. 0.]
         [1. 0.]
         [1. 1.]]
        """
        indices = np.asarray(indices)
        nvecs, alphas = indices[:, :-1], indices[:, -1]
        return self.translate(nvecs, np.array(self.atom_positions)[alphas])

    def estimate_index(self, pos: Union[float, Sequence[float]]) -> np.ndarray:
        """Returns the nearest matching lattice index (n, alpha) for a position.

        Parameters
        ----------
        pos : array_like or float
            The position of the site in world coordinates.

        Returns
        -------
        nvec : np.ndarray
            The estimated translation vector :math:`n`.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom()
        >>> latt.add_connections(1)
        >>> latt.estimate_index([1.2, 0.2])
        [1 0]
        """
        pos = np.asarray(pos)
        n = np.asarray(np.round(self._vectors_inv @ pos, decimals=0), dtype="int")
        return n

    def get_neighbors(
        self, nvec: Union[int, Sequence[int]] = None, alpha: int = 0, distidx: int = 0
    ) -> np.ndarray:
        """Returns the neighour indices of a given site by transforming neighbor data.

        Parameters
        ----------
        nvec: (D) array_like or int, optional
            The translation vector, the default is the origin.
        alpha: int, optional
            The atom index, default is 0.
        distidx: int, optional
            The index of distance to the neighbors, default is 0 (nearest neighbors).

        Returns
        -------
        indices: (N, D+1) np.ndarray
            The lattice indices of the neighbor sites.

        Raises
        ------
        NotAnalyzedError
            Raised if the lattice distances and base-neighbors haven't been computed.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom()
        >>> latt.add_connections(1)
        >>> latt.get_neighbors(nvec=[0, 0], alpha=0, distidx=0)
        [[ 1  0  0]
         [ 0 -1  0]
         [-1  0  0]
         [ 0  1  0]]
        """
        if nvec is None:
            nvec = np.zeros(self.dim)
        self._assert_analyzed()
        logger.debug(
            "Computing neighbor-indices of %s, %i (distidx: %i)", nvec, alpha, distidx
        )

        nvec = np.atleast_1d(nvec)
        keys = list(sorted(self._base_neighbors[alpha].keys()))
        dist = keys[distidx]
        indices = self._base_neighbors[alpha][dist]
        indices_transformed = indices.copy()
        indices_transformed[:, :-1] += nvec.astype(np.int64)
        logger.debug("Neighbour-indices: %s", indices_transformed)

        return indices_transformed

    def get_neighbor_positions(
        self, nvec: Union[int, Sequence[int]] = None, alpha: int = 0, distidx: int = 0
    ) -> np.ndarray:
        """Returns the neighour positions of a given site by transforming neighbor data.

        Parameters
        ----------
        nvec: (D) array_like or int, optional
            The translation vector, the default is the origin.
        alpha: int, optional
           The site index, default is 0.
        distidx: int, default
            The index of distance to the neighbors, default is 0 (nearest neighbors).

        Returns
        -------
        positions: (N, D) np.ndarray
            The positions of the neighbor sites.

        Raises
        ------
        NotAnalyzedError
            Raised if the lattice distances and base-neighbors haven't been computed.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom()
        >>> latt.add_connections(1)
        >>> latt.get_neighbor_positions(nvec=[0, 0], alpha=0, distidx=0)
        [[ 1.  0.]
         [ 0. -1.]
         [-1.  0.]
         [ 0.  1.]]
        """
        if nvec is None:
            nvec = np.zeros(self.dim)
        self._assert_analyzed()
        logger.debug(
            "Computing neighbor-positions of %s, %i (distidx: %i)", nvec, alpha, distidx
        )

        indices = self.get_neighbors(nvec, alpha, distidx)
        nvecs, alphas = indices[:, :-1], indices[:, -1]
        atom_pos = self._positions[alphas]
        positions = self.translate(nvecs, atom_pos)
        logger.debug("Neighbour-positions: %s", positions)

        return positions

    def get_neighbor_vectors(
        self, alpha: int = 0, distidx: int = 0, include_zero: bool = False
    ) -> np.ndarray:
        """Returns the vectors to the neigbor sites of an atom in the unit cell.

        Parameters
        ----------
        alpha : int, optional
            Index of the base atom. The default is the first atom in the unit cell.
        distidx : int, default
            Index of distance to the neighbors, default is 0 (nearest neighbors).
        include_zero : bool, optional
            Flag if zero-vector is included in result. The default is ``False``.

        Returns
        -------
        vectors : np.ndarray
            The vectors from the site of the atom :math:`alpha` to the neighbor sites.

        Raises
        ------
        NotAnalyzedError
            Raised if the lattice distances and base-neighbors haven't been computed.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom()
        >>> latt.add_connections(1)
        >>> latt.get_neighbor_vectors(alpha=0, distidx=0)
        [[ 1.  0.]
         [ 0. -1.]
         [-1.  0.]
         [ 0.  1.]]
        """
        self._assert_analyzed()
        logger.debug(
            "Computing neighbor-vectors of atom %i (distidx: %i)", alpha, distidx
        )

        pos0 = self._positions[alpha]
        pos1 = self.get_neighbor_positions(alpha=alpha, distidx=distidx)
        if include_zero:
            pos1 = np.append(np.zeros((1, self.dim)), pos1, axis=0)
        vecs = pos1 - pos0
        logger.debug("Neighbour-vectors: %s", vecs)

        return vecs

    def fourier_weights(
        self, k: ArrayLike, alpha: int = 0, distidx: int = 0
    ) -> np.ndarray:
        """Returns the Fourier-weight for a given vector.

        Parameters
        ----------
        k : array_like
            The wavevector to compute the lattice Fourier-weights.
        alpha : int, optional
            Index of the base atom. The default is the first atom in the unit cell.
        distidx : int, default
            Index of distance to the neighbors, default is 0 (nearest neighbors).

        Returns
        -------
        weight : np.ndarray
        """
        vecs = self.get_neighbor_vectors(alpha=alpha, distidx=distidx)
        # weights = np.sum([np.exp(1j * np.dot(k, v)) for v in vecs])
        weights = np.sum(np.exp(1j * np.inner(k, vecs)))
        return weights

    def get_base_atom_dict(
        self, atleast2d: bool = True
    ) -> Dict[Any, List[Union[np.ndarray, Any]]]:
        """Returns a dictionary containing the positions for eatch of the base atoms.

        Parameters
        ----------
        atleast2d : bool, optional
            If True, one-dimensional coordinates will be cast to 2D vectors.

        Returns
        -------
        atom_pos : dict
            The positions of the atoms as a dictionary.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom([0, 0], atom="A")
        >>> latt.add_atom([0.5, 0], atom="B")
        >>> latt.add_atom([0.5, 0.5], atom="B")
        >>> latt.get_base_atom_dict()
        {
            Atom(A, radius=0.2, 0): [array([0, 0])],
            Atom(B, radius=0.2, 1): [array([0.5, 0. ]), array([0.5, 0.5])]
        }
        """
        atom_pos = dict()
        for atom, pos in zip(self._atoms, self._positions):
            if atleast2d and self.dim == 1:
                pos = np.array([pos, 0])

            if atom.name in atom_pos.keys():
                atom_pos[atom].append(pos)
            else:
                atom_pos[atom] = [pos]
        return atom_pos

    # noinspection PyShadowingNames
    def check_points(
        self,
        points: np.ndarray,
        shape: Union[int, Sequence[int], AbstractShape],
        relative: bool = False,
        pos: Union[float, Sequence[float]] = None,
        tol: float = 1e-3,
    ) -> np.ndarray:
        """Returns a mask for the points in the given shape.

        Parameters
        ----------
        points: (M, N) np.ndarray
            The points in cartesian coordinates.
        shape: (N) array_like or int or AbstractShape
            shape of finite size lattice to build.
        relative: bool, optional
            If True the shape will be multiplied by the cell size of the model.
            The default is True.
        pos: (N) array_like or int, optional
            Optional position of the section to build. If ``None`` the origin is used.
        tol: float, optional
            The tolerance for checking the points. The default is ``1e-3``.

        Returns
        -------
        mask: (M) np.ndarray
            The mask for the points inside the shape.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> shape = (2, 2)
        >>> points = np.array([[0, 0], [2, 2], [3, 2]])
        >>> latt.check_points(points, shape)
        [ True  True False]
        """
        if isinstance(shape, AbstractShape):
            return shape.contains(points, tol)
        else:
            shape = np.atleast_1d(shape)
            if len(shape) != self.dim:
                raise ValueError(
                    f"Dimension of shape {len(shape)} doesn't "
                    f"match the dimension of the lattice {self.dim}"
                )
            if relative:
                shape += np.max(self.vectors, axis=0) - 0.1 * self.norms

            pos = np.zeros(self.dim) if pos is None else np.array(pos, dtype=np.float64)
            pos -= tol
            end = pos + shape + tol

            mask = (pos[0] <= points[:, 0]) & (points[:, 0] <= end[0])
            for i in range(1, self.dim):
                mask = mask & (pos[i] <= points[:, i]) & (points[:, i] <= end[i])
            return mask

    def build_translation_vectors(
        self,
        shape: Union[int, Sequence[int], AbstractShape],
        primitive: bool = False,
        pos: Union[float, Sequence[float]] = None,
        check: bool = True,
        dtype: Union[int, np.dtype] = None,
        oversample: float = 0.0,
    ) -> np.ndarray:
        """Constructs the translation vectors :math:`n` in a given shape.

        Raises
        ------
        ValueError
            Raised if the dimension of the position doesn't match the dimension of
            the lattice.

        Parameters
        ----------
        shape: (N) array_like or int
            shape of finite size lattice to build.
        primitive: bool, optional
            If True the shape will be multiplied by the cell size of the model.
            The default is True.
        pos: (N) array_like or int, optional
            Optional position of the section to build. If ``None`` the origin is used.
        check: bool, optional
            If `True` the positions of the translation vectors are checked and
            filtered. The default is True. This should only be disabled if
            filtered later.
        dtype: int or np.dtype, optional
            Optional data-type for storing the lattice indices. By default, the given
            limits are checked to determine the smallest possible data-type.
        oversample: float, optional
            Faktor for upscaling limits for initial index grid. This ensures that all
            positions are included. Only needed if corner points are missing.
            The default is 0.

        Returns
        -------
        nvecs: (M, N) np.ndarray
            The translation-vectors in lattice-coordinates.

        Examples
        --------
        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.build_translation_vectors((2, 2))
        [[0 0]
         [0 1]
         [0 2]
         [1 0]
         [1 1]
         [1 2]
         [2 0]
         [2 1]
         [2 2]]
        """
        # Build lattice indices
        if isinstance(shape, AbstractShape):
            pos, stop = shape.limits().T
            shape = stop - pos
        shape = np.atleast_1d(shape)
        if len(shape) != self.dim:
            raise ValueError(
                f"Dimension of shape {len(shape)} doesn't "
                f"match the dimension of the lattice {self.dim}"
            )
        logger.debug("Building nvecs: %s at %s", shape, pos)

        if primitive:
            shape = np.array(shape) * np.max(self.vectors, axis=0) - 0.1 * self.norms
        if pos is None:
            pos = np.zeros(self.dim)
        end = pos + shape

        # Estimate the maximum needed translation vector to reach all points
        max_nvecs = np.array(
            [self.itranslate(pos)[0], self.itranslate(end)[0]], dtype=np.float64
        )
        for i in range(1, self.dim):
            for idx in itertools.combinations(range(self.dim), r=i):
                _pos = end.copy()
                _pos[np.array(idx)] = 0
                index = self.itranslate(_pos)[0]
                max_nvecs[0] = np.min([index, max_nvecs[0]], axis=0)
                max_nvecs[1] = np.max([index, max_nvecs[1]], axis=0)
        # Pad maximum translation vectors and create index limits
        padding = oversample * shape + 1
        max_nvecs += [-padding, +padding]
        limits = max_nvecs.astype(np.int64).T
        logger.debug("Limits: %s, %s", limits[:, 0], limits[:, 1])

        # Generate translation vectors with too many points to reach each corner
        nvecs = vindices(limits, sort_axis=0, dtype=dtype)
        logger.debug("%s Translation vectors built", len(nvecs))
        if check:
            logger.debug("Filtering nvec's")
            # Filter points in the given volume
            positions = np.dot(nvecs, self.vectors[np.newaxis, :, :])[:, 0, :]
            mask = self.check_points(positions, shape, primitive, pos)
            nvecs = nvecs[mask]
        return nvecs

    # noinspection PyShadowingNames
    def build_indices(
        self,
        shape: Union[int, Sequence[int], AbstractShape],
        primitive: bool = False,
        pos: Union[float, Sequence[float]] = None,
        check: bool = True,
        callback: Callable = None,
        dtype: Union[int, str, np.dtype] = None,
        return_pos: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Constructs the lattice indices .math:`(n, α)` in the given shape.

        Raises
        ------
        ValueError
            Raised if the dimension of the position doesn't match the dimension
            of the lattice.

        Parameters
        ----------
        shape: (N) array_like or int or AbstractShape
            shape of finite size lattice to build.
        primitive: bool, optional
            If True the shape will be multiplied by the cell size of the model.
            The default is True.
        pos: (N) array_like or int, optional
            Optional position of the section to build. If ``None`` the origin is used.
        check : bool, optional
            If True the positions of the translation vectors are checked and
            filtered. The default is True. This should only be disabled if
            filtered later.
        callback: callable, optional
            Optional callable for filtering sites.
            The indices and positions are passed as arguments.
        dtype: int or str or np.dtype, optional
            Optional data-type for storing the lattice indices. By default, the given
            limits are checked to determine the smallest possible data-type.
        return_pos: bool, optional
            Flag if positions should be returned with the indices. This can speed up
            the building process, since the positions have to be computed here anyway.
            The default is ``False``.

        Returns
        -------
        indices: (M, N+1) np.ndarray
            The lattice indices of the sites in the format
            .math:`(n_1, ..., n_d, α)`.
        positions: (M, N) np.ndarray
            Corresponding positions. Only returned if ``return_positions=True``.

        Examples
        --------
        Build indices of a linear chain with two atoms in the unit cell:

        >>> latt = LatticeStructure(np.eye(2))
        >>> latt.add_atom([0.0, 0.0], "A")
        >>> latt.add_atom([0.5, 0.5], "B")
        >>> indices, positions = latt.build_indices((2, 1), return_pos=True)

        The indices contain the translation vector and the atom index

        >>> indices
        [[0 0 0]
         [0 0 1]
         [0 1 0]
         [1 0 0]
         [1 0 1]
         [1 1 0]
         [2 0 0]
         [2 1 0]]

        The positions are the positions of the atoms in the same order of the indices:

        >>> positions
        [[0.  0. ]
         [0.5 0.5]
         [0.  1. ]
         [1.  0. ]
         [1.5 0.5]
         [1.  1. ]
         [2.  0. ]
         [2.  1. ]]
        """
        logger.debug("Building lattice-indices: %s at %s", shape, pos)

        # Build lattice indices
        nvecs = self.build_translation_vectors(shape, primitive, pos, False, dtype)
        ones = np.ones(nvecs.shape[0], dtype=nvecs.dtype)
        arrays = [np.c_[nvecs, i * ones] for i in range(self.num_base)]
        cols = self.dim + 1
        indices = np.ravel(arrays, order="F")
        indices = indices.reshape(cols, int(indices.shape[0] / cols)).T

        logger.debug("Computing positions of sub-lattices")
        # Compute positions for filtering
        positions = [self.translate(nvecs, pos) for pos in self.atom_positions]
        positions = interweave(positions)

        if check:
            # Filter points in the given volume
            logger.debug("Filtering points")
            mask = self.check_points(positions, shape, primitive, pos)
            indices = indices[mask]
            positions = positions[mask]

        # Filter points with user method
        if callback is not None:
            logger.debug("Applying callback-method")
            mask = callback(indices, positions)
            indices = indices[mask]
            positions = positions[mask]

        logger.debug("Created %i lattice sites", len(indices))
        return indices, positions if return_pos else None

    def _filter_neighbors(self, indices, neighbors, distances, x_ind=None):
        logger.debug("Filtering neighbors")

        x_ind = indices if x_ind is None else x_ind

        # Add dummy index for invalid index
        invalid_ind = len(indices)
        indices_padded = np.append(indices, -np.ones((1, indices.shape[1])), axis=0)
        alphas = indices_padded[neighbors, -1]

        # Remove not connected neighbors
        dist_matrix = self._distance_matrix
        num_base = len(dist_matrix)
        for a1 in range(num_base):
            mask1 = (x_ind[:, -1] == a1)[:, None]
            for a2 in range(num_base):
                connected = np.isin(distances, dist_matrix[a1][a2])
                mask = mask1 & (alphas == a2) & (~connected)
                neighbors[mask] = invalid_ind
                distances[mask] = np.inf

        logger.debug("Re-sorting and cleaning up neighbors")

        # Resort neighbors
        i = np.arange(len(distances))[:, np.newaxis]
        j = np.argsort(distances, axis=1)
        distances = distances[i, j]
        neighbors = neighbors[i, j]
        num_sites = len(neighbors)

        if not np.any(distances == np.inf):
            # No invalid entries found. This usually happens for (2, 2, ..) systems
            # which results in a bug for the neighbor data:
            # Assume a 1D chain of 3 atoms. The outer two atoms each have only one
            # neighbor (the center atom), whereas the center atom has two.
            # In a model of two atoms no atom has two neighbors, which prevents
            # the algorithm to create an array of size (N, 2).
            # To prevent this a column of invalid values is appended to the data.
            # Since the system size is small this doesn't create any memory issues.
            shape = num_sites, 1
            neighbors = np.append(neighbors, np.full(shape, invalid_ind), axis=1)
            distances = np.append(distances, np.full(shape, np.inf), axis=1)
        else:
            # Remove columns containing only invalid data
            all_valid = np.any(distances != np.inf, axis=0)
            distances = distances[:, all_valid]
            neighbors = neighbors[:, all_valid]

        return neighbors, distances

    # noinspection PyShadowingNames
    def compute_neighbors(
        self, indices: ArrayLike, positions: ArrayLike, num_jobs: int = 1
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Computes the neighbors for the given points.

        Parameters
        ----------
        indices : (N, D+1) array_like
            The lattice indices of the sites in the format
            .math:`(n_1, ..., n_D, α)` where N is the number of sites and D the
            dimension of the lattice.
        positions : (N, D) array_like
            The positions of the sites in cartesian coordinates where N is the number
            of sites and D the dimension of the lattice.
        num_jobs : int, optional
            Number of jobs to schedule for parallel processing.
            If ``-1`` is given all processors are used. The default is ``1``.

        Returns
        -------
        neighbors: (N, M) np.ndarray
            The indices of the neighbors in ``positions``. M is the maximum
            number of neighbors previously computed in the ``analyze`` method.
        distances: (N, M) np.ndarray
            The corresponding distances of the neighbors.

        See Also
        --------
        analyze : Used to pre-compute the base neighbors of the unit cell.

        Examples
        --------
        Construct indices of a one dimensional lattice:

        >>> latt = LatticeStructure(1)
        >>> latt.add_atom()
        >>> latt.add_connections()
        >>> indices, positions = latt.build_indices(3, return_pos=True)
        >>> positions
        [[0.]
         [1.]
         [2.]
         [3.]]

        Compute the neighbors of the constructed sites

        >>> neighbors, distances = latt.compute_neighbors(indices, positions)
        >>> neighbors
        [[1 4]
         [2 0]
         [3 1]
         [2 4]]

        >>> indices
        [[ 1. inf]
         [ 1.  1.]
         [ 1.  1.]
         [ 1. inf]]

        The neighbor indices and distances of sites with less than the maximum number
        of neighbors are filled up with an invalid index (here: 4) and ``np.inf``
        as distance.
        """

        logger.debug("Querying neighbors of %i points", len(positions))

        # Set neighbor query parameters and build tree
        k = np.sum(np.sum(self._raw_num_neighbors, axis=1)) + 1
        max_dist = np.max(self.distances) + 0.1 * np.min(self._raw_distance_matrix)
        tree = KDTree(positions, k=k, max_dist=max_dist)
        logger.debug("Max. number of neighbors: %i", k)
        logger.debug("Max. neighbor distance:   %f", max_dist)
        # Query and filter neighbors
        neighbors, distances = tree.query(
            num_jobs=num_jobs, decimals=self.DIST_DECIMALS
        )
        neighbors, distances = self._filter_neighbors(indices, neighbors, distances)

        return neighbors, distances

    def copy(self) -> "LatticeStructure":
        """LatticeStructure : Creates a (deep) copy of the lattice instance."""
        return deepcopy(self)

    def todict(self) -> dict:
        """Creates a dictionary containing the information of the lattice instance.

        Returns
        -------
        d : dict
            The information defining the current instance.
        """
        d = dict()
        d["vectors"] = self.vectors
        d["atoms"] = self._atoms
        d["positions"] = self._positions
        d["connections"] = self._connections
        return d

    @classmethod
    def fromdict(cls, d):
        """Creates a new instance from information stored in a dictionary.

        Parameters
        ----------
        d : dict
            The information defining the current instance.

        Returns
        -------
        latt : LatticeStructure
            The restored lattice instance.
        """
        self = cls(d["vectors"])
        for pos, at in zip(d["positions"], d["atoms"]):
            self.add_atom(pos, at)
        self._connections = d["connections"]
        self.analyze()
        return self

    def dumps(self):  # pragma: no cover
        """Creates a string containing the information of the lattice instance.

        Returns
        -------
        s : str
            The information defining the current instance.
        """
        lines = list()
        for key, values in self.todict().items():
            head = key + ":"
            lines.append(f"{head:<15}" + "; ".join(str(x) for x in values))
        return "\n".join(lines)

    def dump(self, file: Union[str, int, bytes]) -> None:  # pragma: no cover
        """Save the data of the ``LatticeStructure`` instance.

        Parameters
        ----------
        file : str or int or bytes
            File name to store the lattice. If ``None`` the hash of the lattice is used.
        """
        if file is None:
            file = f"{self.__hash__()}.latt"
        with open(file, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(
        cls, file: Union[str, int, bytes]
    ) -> "LatticeStructure":  # pragma: no cover
        """Load data of a saved ``LatticeStructure`` instance.

        Parameters
        ----------
        file : str or int or bytes
            File name to load the lattice.

        Returns
        -------
        latt : LatticeStructure
            The lattice restored from the file content.
        """
        with open(file, "rb") as f:
            latt = pickle.load(f)
        return latt

    def __hash__(self):
        import hashlib

        sha = hashlib.md5(self.dumps().encode("utf-8"))
        return int(sha.hexdigest(), 16)

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def plot_cell(
        self,
        lw: float = None,
        alpha: float = 0.5,
        cell_lw: float = None,
        cell_ls: str = "--",
        margins: Union[Sequence[float], float] = 0.1,
        legend: bool = None,
        grid: bool = False,
        show_cell: bool = True,
        show_vecs: bool = True,
        show_neighbors: bool = True,
        con_colors: Sequence = None,
        adjustable: str = "box",
        ax: Union[plt.Axes, Axes3D] = None,
        show: bool = False,
    ) -> Union[plt.Axes, Axes3D]:  # pragma: no cover
        """Plot the unit cell of the lattice.

        Parameters
        ----------
        lw : float, optional
            Line width of the neighbor connections.
        alpha : float, optional
            The alpha value of the neighbor sites.
        cell_lw : float, optional
            The line width used for plotting the unit cell outlines.
        cell_ls : str, optional
            The line style used for plotting the unit cell outlines.
        margins : Sequence[float] or float, optional
            The margins of the plot.
        legend : bool, optional
            Flag if legend is shown.
        grid : bool, optional
            If True, draw a grid in the plot.
        show_neighbors : bool, optional
            If True the neighbors are plotted.
        show_vecs : bool, optional
            If True the first unit-cell is drawn.
        show_cell : bool, optional
            If True the outlines of the unit cell are plotted.
        con_colors : Sequence[tuple], optional
            list of colors to override the defautl connection color. Each element
            has to be a tuple with the first two elements being the atom indices of
            the pair and the third element the color, for example ``[('A', 'A', 'r')]``.
        adjustable : None or {'box', 'datalim'}, optional
            If not None, this defines which parameter will be adjusted to meet
            the equal aspect ratio. If 'box', change the physical dimensions of
            the Axes. If 'datalim', change the x or y data limits.
            Only applied to 2D plots.
        ax : plt.Axes or plt.Axes3D or None, optional
            Parent plot. If None, a new plot is initialized.
        show : bool, optional
            If True, show the resulting plot.
        """
        logger.debug("Plotting unit cell")
        if self.dim > 3:
            raise ValueError(f"Plotting in {self.dim} dimensions is not supported!")

        hopz, atomz = range(2)

        fig, ax = subplot(self.dim, adjustable, ax)

        # Draw unit vectors and the cell they spawn.
        if show_vecs:
            vectors = self.vectors
            draw_unit_cell(
                ax, vectors, show_cell, color="k", lw=cell_lw, ls=cell_ls, zorder=hopz
            )
        # Draw sites
        colors = list()
        for i in range(self.num_base):
            atom = self.atoms[i]
            col = atom.color or f"C{i}"
            points = self.atom_positions[i]
            rad = atom.radius
            draw_sites(ax, points, rad, color=col, label=atom.name, zorder=atomz)
            colors.append(col)
        # Draw Neighbors and connections
        ccolor = "k"
        hop_colors = connection_color_array(self.num_base, ccolor, con_colors)
        if show_neighbors:
            position_arr = [list() for _ in range(self.num_base)]
            for i in range(self.num_base):
                p1 = self.atom_positions[i]
                for distidx in range(self.num_distances):
                    try:
                        indices = self.get_neighbors(alpha=i, distidx=distidx)
                        positions = self.get_positions(indices)
                        for idx, p2 in zip(indices, positions):
                            j = idx[-1]
                            col = hop_colors[i][j]
                            draw_vectors(
                                ax, p2 - p1, pos=p1, zorder=hopz, color=col, lw=lw
                            )
                            if np.any(idx[:-1]):
                                position_arr[j].append(p2)
                    except IndexError:
                        pass
            for i in range(self.num_base):
                atom = self.get_atom(i)
                positions = position_arr[i]
                if positions:
                    pos = np.unique(positions, axis=0)
                    rad = 0.6 * atom.radius
                    col = colors[i]
                    draw_sites(ax, pos, rad, color=col, alpha=alpha, zorder=atomz)
        # Configure legend
        if legend is None:
            legend = self.num_base > 1
        if legend:
            ax.legend()
        # Configure grid
        if grid and self.dim < 3:
            ax.set_axisbelow(True)
            ax.grid(b=True, which="major")
        # Adjust margin
        if isinstance(margins, float):
            margins = [margins] * self.dim
        ax.margins(*margins)

        fig.tight_layout()
        if show:
            plt.show()
        return ax

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"dim: {self.dim}, "
            f"num_base: {self.num_base}, "
            f"num_neighbors: {self.num_neighbors})"
        )
